from copy import copy
from typing import Optional

import numpy as np
from scipy import stats
from scipy.special import boxcox, inv_boxcox
from statsmodels.tsa.api import STLForecast
from statsmodels.tsa.arima.model import ARIMA

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import ts_to_table
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.utilities.ts_gapfilling import SimpleGapFiller


class ARIMAImplementation(ModelImplementation):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.arima = None
        self.lambda_value = None
        self.scope = None
        self.actual_ts_len = None

    def fit(self, input_data):
        """ Class fit arima model on data

        :param input_data: data with features, target and ids to process
        """
        source_ts = np.array(input_data.features)
        # Save actual time series length
        self.actual_ts_len = len(source_ts)

        # Apply box-cox transformation for positive values
        transformed_ts = self._apply_boxcox(source_ts)

        # Set parameters
        p = int(self.params.get('p'))
        d = int(self.params.get('d'))
        q = int(self.params.get('q'))
        params = {'order': (p, d, q)}

        self.arima = ARIMA(transformed_ts, **params).fit()

        return self.arima

    def predict(self, input_data: InputData):
        """ Method for time series prediction on forecast length

        :param input_data: data with features, target and ids to process

        :return output_data: output data with smoothed time series
        """
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length
        self.handle_new_data(input_data)

        start_id = self.actual_ts_len
        end_id = start_id + forecast_length - 1

        predicted = self.arima.predict(start=start_id,
                                       end=end_id)
        predicted = self._inverse_boxcox(predicted=predicted,
                                         lambda_param=self.lambda_value)
        predict = self._inverse_shift(predicted)

        predict = np.array(predict).reshape(1, -1)
        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)

        return output_data

    def predict_for_fit(self, input_data: InputData):
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length
        fitted_values = self.arima.fittedvalues

        fitted_values = self._inverse_boxcox(predicted=fitted_values,
                                             lambda_param=self.lambda_value)
        fitted_values = self._inverse_shift(fitted_values)

        diff = int(self.actual_ts_len - len(fitted_values))
        # If first elements skipped
        if diff != 0:
            # Fill nans with first values
            first_element = fitted_values[0]
            first_elements = [first_element] * diff
            first_elements.extend(list(fitted_values))

            fitted_values = np.array(first_elements)

        _, predict = ts_to_table(idx=input_data.idx,
                                 time_series=fitted_values,
                                 window_size=forecast_length)

        new_idx, target_columns = ts_to_table(idx=input_data.idx,
                                              time_series=input_data.target,
                                              window_size=forecast_length)
        input_data.idx = new_idx
        input_data.target = target_columns

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)

        return output_data

    def handle_new_data(self, input_data: InputData):
        """
        Method to update x samples inside a model (used when we want to use old model to a new data)

        :param input_data: new input_data
        """
        if input_data.idx[0] > self.actual_ts_len:
            self.arima = self.fit(input_data)
            self.log.info("Arima refitted for handling a new data")

    def _apply_boxcox(self, source_ts):
        min_value = np.min(source_ts)
        if min_value > 0:
            pass
        else:
            # Making a shift to positive values
            self.scope = abs(min_value) + 1
            source_ts = source_ts + self.scope

        _, self.lambda_value = stats.boxcox(source_ts)
        transformed_ts = boxcox(source_ts, self.lambda_value)

        return transformed_ts

    def _inverse_boxcox(self, predicted, lambda_param):
        """ Method apply inverse Box-Cox transformation """
        if lambda_param == 0:
            return np.exp(predicted)
        else:
            res = inv_boxcox(predicted, lambda_param)
            res = self._filling_gaps(res)
            return res

    def _inverse_shift(self, values):
        """ Method apply inverse shift operation """
        if self.scope is None:
            pass
        else:
            values = values - self.scope

        return values

    @staticmethod
    def _filling_gaps(res):
        nan_ind = np.argwhere(np.isnan(res))
        res[nan_ind] = -100.0

        # Gaps in first and last elements fills with mean value
        if 0 in nan_ind:
            res[0] = np.mean(res)
        if int(len(res) - 1) in nan_ind:
            res[int(len(res) - 1)] = np.mean(res)

        # Gaps in center of timeseries fills with linear interpolation
        if len(np.ravel(np.argwhere(np.isnan(res)))) != 0:
            gf = SimpleGapFiller()
            res = gf.linear_interpolation(res)

        return res


class STLForecastARIMAImplementation(ModelImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.model = None
        self.lambda_param = None
        self.scope = None
        self.actual_ts_len = None

    def fit(self, input_data):
        """ Class fit STLForecast arima model on data

        :param input_data: data with features, target and ids to process
        """

        source_ts = np.array(input_data.features)
        # Save actual time series length
        self.actual_ts_len = len(source_ts)

        p = int(self.params.setdefault('p', 2))
        d = int(self.params.setdefault('d', 0))
        q = int(self.params.setdefault('q', 2))
        period = int(self.params.setdefault('period', 365))
        params = {'period': period, 'model_kwargs': {'order': (p, d, q)}}
        self.model = STLForecast(source_ts, ARIMA, **params).fit()

        return self.model

    def predict(self, input_data: InputData) -> OutputData:
        """ Method for time series prediction on forecast length

        :param input_data: data with features, target and ids to process

        :return output_data: output data with smoothed time series
        """
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length

        # in case in(out) sample forecasting
        self.handle_new_data(input_data)
        start_id = self.actual_ts_len
        end_id = start_id + forecast_length - 1
        predicted = self.model.get_prediction(start=start_id, end=end_id).predicted_mean

        predict = np.array(predicted).reshape(1, -1)
        new_idx = np.arange(start_id, end_id + 1)

        input_data.idx = new_idx

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        idx = input_data.idx
        target = input_data.target
        fitted_values = self.model.get_prediction(start=idx[0], end=idx[-1]).predicted_mean
        diff = int(self.actual_ts_len) - len(fitted_values)
        # If first elements skipped
        if diff != 0:
            # Fill nans with first values
            first_element = fitted_values[0]
            first_elements = [first_element] * diff
            first_elements.extend(list(fitted_values))

            fitted_values = np.array(first_elements)

        _, predict = ts_to_table(idx=idx,
                                 time_series=fitted_values,
                                 window_size=forecast_length)

        new_idx, target_columns = ts_to_table(idx=idx,
                                              time_series=target,
                                              window_size=forecast_length)

        input_data.idx = new_idx
        input_data.target = target_columns
        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def handle_new_data(self, input_data: InputData):
        """ Refit model if use new test data"""
        if input_data.idx[0] > self.actual_ts_len:
            self.model = self.fit(input_data)
            self.log.info("STL Arima refitted for handling a new data")



from abc import ABC, abstractmethod


class GraphOperationRepository(ABC):
    """ Base repository in order to extract suitable models for each of graph structure
    from specific files/configs etc """

    @abstractmethod
    def get_operations(self, **kwargs):
        """ Get models by specified model keys """
        raise NotImplementedError()

    @abstractmethod
    def get_all_operations(self):
        """ Get all models with all keys """
        raise NotImplementedError()



import json
import os
import shutil
import time
from datetime import datetime, timedelta
from typing import List, Optional

import requests

from fedot.core.pipelines.pipeline import Pipeline
from fedot.remote.infrastructure.clients.client import Client

USER_TOKEN_KEY = 'x-jwt-auth'
GROUP_TOKEN_KEY = 'x-jwt-models-execution'

DEFAULT_EXEC_PARAMS = {
    'container_input_path': "/home/FEDOT/input_data_dir",
    'container_output_path': "/home/FEDOT/output_data_dir",
    'container_config_path': "/home/FEDOT/.config",
    'container_image': "fedot:dm-9",
    'timeout': 360
}

# example of connection params for DataMall
DEFAULT_CONNECT_PARAMS = {
    'FEDOT_LOGIN': 'fedot',
    'FEDOT_PASSWORD': 'fedot-password',
    'AUTH_SERVER': 'http://10.32.0.51:30880/b',
    'CONTR_SERVER': 'http://10.32.0.51:30880/models-controller',
    'PROJECT_ID': '83',
    'DATA_ID': '60'
}


# TO BE MOVED TO PYPI AS EXTERNAL LIB

class DataMallClient(Client):
    def __init__(self, connect_params: dict, exec_params: dict, output_path: Optional[str] = None):
        authorization_server = connect_params['AUTH_SERVER']
        controller_server = connect_params['CONTR_SERVER']
        self.authorization_server = os.environ['AUTH_SERVER'] if authorization_server is None else authorization_server
        self.controller_server = os.environ['CONTR_SERVER'] if controller_server is None else controller_server
        self._user_token = None
        self.user = None
        self._current_project_id = None
        self.group_token = None

        self._login(login=connect_params['FEDOT_LOGIN'],
                    password=connect_params['FEDOT_PASSWORD'])

        pid = connect_params['PROJECT_ID']
        group = self._create_execution_group(project_id=pid)
        self._set_group_token(project_id=pid, group_id=group['id'])

        super().__init__(connect_params, exec_params, output_path)

    def create_task(self, config) -> str:
        data_id = self.connect_params['DATA_ID']
        created_ex = self._create_execution(f"{self.exec_params['container_input_path']}",
                                            self.exec_params['container_output_path'],
                                            self.exec_params['container_config_path'],
                                            self.exec_params['container_image'],
                                            self.exec_params['timeout'],
                                            config=config)
        return created_ex['id']

    def wait_until_ready(self) -> timedelta:
        statuses = ['']
        all_executions = self._get_executions()
        self._logger.info(all_executions)
        start = datetime.now()
        while any(s not in ['Succeeded', 'Failed', 'Timeout', 'Interrupted'] for s in statuses):
            executions = self._get_executions()
            statuses = [execution['status'] for execution in executions]
            self._logger.info([f"{execution['id']}={execution['status']};" for execution in executions])
            time.sleep(5)
        end = datetime.now()
        ex_time = end - start
        return ex_time

    def _login(self, login: str, password: str) -> None:
        response = requests.post(
            url=f'{self.authorization_server}/users/login',
            json={
                'login': login,
                'password': password
            }
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to get user token. Reason: {response.text}')

        self._user_token = response.cookies['x-jwt-auth']
        self.user = json.loads(response.text)

    def _set_group_token(self, project_id: int, group_token: str = None, group_id: int = None) -> None:
        if group_token is not None:
            self.group_token = group_token
            self._current_project_id = project_id
            return

        if group_id is not None:
            group = self._get_execution_group(
                project_id=project_id,
                group_id=group_id
            )

            self.group_token = group['token']
            self._current_project_id = group['project_id']

            return

        raise ValueError(f'You have to specify project_id and token/group_id!')

    def _get_execution_groups(self, project_id: int) -> List[dict]:
        response = requests.get(
            url=f'{self.controller_server}/execution-groups/{project_id}',
            cookies={USER_TOKEN_KEY: self._user_token}
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to get execution groups. Reason: {response.text}')

        return json.loads(response.text)

    def _get_execution_group(self, project_id: int, group_id: int) -> dict:
        response = requests.get(
            url=f'{self.controller_server}/execution-groups/{project_id}/{group_id}',
            cookies={USER_TOKEN_KEY: self._user_token}
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to get execution group. Reason: {response.text}')

        return json.loads(response.text)

    def _create_execution_group(self, project_id: int) -> dict:
        response = requests.post(
            url=f'{self.controller_server}/execution-groups/{project_id}',
            cookies={USER_TOKEN_KEY: self._user_token}
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to create execution group. Reason: {response.text}')

        return json.loads(response.text)

    def _get_executions(self):
        response = requests.get(
            url=f'{self.controller_server}/executions/{self._current_project_id}',
            cookies={USER_TOKEN_KEY: self._user_token},
            headers={GROUP_TOKEN_KEY: self.group_token}
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to get executions. Reason: {response.text}')

        return json.loads(response.text)

    def _get_execution(self, execution_id: int):
        response = requests.get(
            url=f'{self.controller_server}/executions/{self._current_project_id}/{execution_id}',
            cookies={USER_TOKEN_KEY: self._user_token},
            headers={GROUP_TOKEN_KEY: self.group_token}
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to get execution. Reason: {response.text}')

        return json.loads(response.text)

    def _create_execution(self, container_input_path: str,
                          container_output_path: str,
                          container_config_path: str,
                          container_image: str,
                          timeout: int,
                          config: bytes) -> dict:
        response = requests.post(
            url=f'{self.controller_server}/executions/{self._current_project_id}',
            cookies={USER_TOKEN_KEY: self._user_token},
            headers={GROUP_TOKEN_KEY: self.group_token},
            files={
                'input_path': (None, container_input_path),
                'output_path': (None, container_output_path),
                'config_path': (None, container_config_path),
                'image': (None, container_image),
                'timeout': (None, timeout),
                'config_file': ('config', config)
            }
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to create execution. Reason: {response.text}')

        return json.loads(response.text)

    def _stop_execution(self, execution_id: int) -> None:
        response = requests.delete(
            url=f'{self.controller_server}/executions/{self._current_project_id}/{execution_id}',
            cookies={USER_TOKEN_KEY: self._user_token},
            headers={GROUP_TOKEN_KEY: self.group_token}
        )

        if response.status_code != 204:
            raise ValueError(f'Unable to stop execution. Reason: {response.text}')

    def download_result(self, execution_id: int, result_cls=Pipeline) -> Pipeline:
        response = requests.get(
            url=f'{self.controller_server}/executions/{self._current_project_id}/{execution_id}/download',
            cookies={USER_TOKEN_KEY: self._user_token},
            headers={GROUP_TOKEN_KEY: self.group_token},
            stream=True
        )

        if response.status_code != 200:
            raise ValueError(f'Unable to download results. Reason: {response.text}')

        tmp_path = f'_tmp_{int(datetime.utcnow().timestamp() * 1000)}'
        try:
            with open(tmp_path, 'wb') as tmp_file:
                shutil.copyfileobj(response.raw, tmp_file)
            shutil.unpack_archive(tmp_path, f'{self.output_path}/execution-{execution_id}', 'zip')
        finally:
            try:
                os.remove(tmp_path)
            except FileNotFoundError:
                pass

        results_path_out = os.path.join(self.output_path,
                                        f'execution-{execution_id}',
                                        'out')
        results_folder = os.listdir(results_path_out)[0]
        load_path = os.path.join(results_path_out, results_folder, 'fitted_pipeline.json')
        pipeline = result_cls.from_serialized(load_path)

        clean_dir(results_path_out)
        return pipeline


def clean_dir(results_path_out):
    for root, dirs, files in os.walk(results_path_out):
        for file in files:
            os.remove(os.path.join(root, file))



from __future__ import annotations

import os
from functools import partial
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from fedot.core.data.data import process_target_and_features, get_indices_from_file, array_to_input_data
from fedot.core.data.data_detection import TextDataDetector, TimeSeriesDataDetector
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum


class MultiModalData(dict):
    """ Dictionary with InputData as values and primary node names as keys """

    def __init__(self, *arg, **kw):
        super(MultiModalData, self).__init__(*arg, **kw)

        # Check if input data contains different targets
        self.contain_side_inputs = not all(value.supplementary_data.is_main_target for value in self.values())

    @property
    def idx(self):
        for input_data in self.values():
            if input_data.supplementary_data.is_main_target:
                return input_data.idx

    @property
    def task(self):
        for input_data in self.values():
            if input_data.supplementary_data.is_main_target:
                return input_data.task

    @task.setter
    def task(self, value):
        """ Update task for all input data """
        for input_data in self.values():
            input_data.task = value

    @property
    def target(self):
        """ Return main target from InputData blocks """
        for input_data in self.values():
            if input_data.supplementary_data.is_main_target:
                return input_data.target

    @target.setter
    def target(self, value):
        """ Update target for all input data """
        for input_data in self.values():
            input_data.target = value

    @property
    def data_type(self):
        return [input_data.data_type for input_data in iter(self.values())]

    @property
    def num_classes(self) -> Optional[int]:
        if self.task.task_type == TaskTypesEnum.classification:
            return len(np.unique(self.target))
        else:
            return None

    @property
    def class_labels(self) -> Optional[List[Union[int, str, float]]]:
        if self.task.task_type == TaskTypesEnum.classification and self.target is not None:
            return np.unique(self.target)
        else:
            return None

    def shuffle(self):
        # TODO implement multi-modal shuffle
        pass

    def extract_data_source(self, source_name):
        """
            Function for extraction data_source from MultiModalData
            :param source_name: string with user-specified name of source
            :return target_data: selected source InputData
        """
        full_target_name = [key for key, _ in self.items() if source_name == key.split('/')[-1]][0]
        source_data = self[full_target_name]
        return source_data

    def subset_range(self, start: int, end: int):
        for key in self.keys():
            self[key] = self[key].subset_range(start, end)
        return self

    def subset_indices(self, selected_idx: List):
        for key in self.keys():
            self[key] = self[key].subset_indices(selected_idx)
        return self

    @classmethod
    def from_csv_time_series(cls,
                             task: Union[Task, str] = 'ts_forecasting',
                             file_path=None,
                             delimiter=',',
                             is_predict=False,
                             var_names=None,
                             target_column: Optional[str] = '',
                             idx_column: Optional[str] = 'datetime') -> MultiModalData:
        ts_data_detector = TimeSeriesDataDetector()
        df = pd.read_csv(file_path, sep=delimiter)
        idx = get_indices_from_file(df, file_path, idx_column)
        if isinstance(task, str):
            task = Task(TaskTypesEnum(task))
        if not var_names:
            var_names = list(set(df.columns) - set(idx_column))

        if is_predict:
            raise NotImplementedError(
                'Multivariate predict not supported in this function yet.')
        else:
            data = ts_data_detector.prepare_multimodal_data(dataframe=df,
                                                            columns=var_names)

            if target_column is not None:
                target = np.array(df[target_column])
            else:
                target = np.array(df[df.columns[-1]])

            # create labels for data sources
            data_part_transformation_func = partial(array_to_input_data,
                                                    idx=idx, target_array=target, task=task,
                                                    data_type=DataTypesEnum.ts)

            sources = dict((ts_data_detector.new_key_name(data_part_key),
                            data_part_transformation_func(features_array=data_part))
                           for (data_part_key, data_part) in data.items())
            multi_modal_data = MultiModalData(sources)

        return multi_modal_data

    @classmethod
    def from_csv(cls,
                 file_path: Optional[Union[os.PathLike, str]] = None,
                 delimiter=',',
                 task: Union[Task, str] = 'classification',
                 text_columns: Optional[Union[str, List[str]]] = None,
                 columns_to_drop: Optional[List[str]] = None,
                 target_columns: Union[str, List[str]] = '',
                 index_col: Optional[Union[str, int]] = 0) -> MultiModalData:
        """
        :param file_path: the path to the CSV with data
        :param columns_to_drop: the names of columns that should be dropped
        :param delimiter: the delimiter to separate the columns
        :param task: the task that should be solved with data
        :param text_columns: names of columns that contain text data
        :param target_columns: name of target column (last column if empty and no target if None)
        :param index_col: column name or index to use as the Data.idx;
            if None then arrange new unique index
        :return: MultiModalData object with text and table data sources as InputData
        """

        text_data_detector = TextDataDetector()
        data_frame = pd.read_csv(file_path, sep=delimiter, index_col=index_col)
        if columns_to_drop:
            data_frame = data_frame.drop(columns_to_drop, axis=1)

        idx = data_frame.index.to_numpy()
        if isinstance(task, str):
            task = Task(TaskTypesEnum(task))
        text_columns = [text_columns] if isinstance(text_columns, str) else text_columns

        if not text_columns:
            text_columns = text_data_detector.define_text_columns(data_frame)

        data_text = text_data_detector.prepare_multimodal_data(data_frame, text_columns)
        data_frame_table = data_frame.drop(columns=text_columns)
        table_features, target = process_target_and_features(data_frame_table, target_columns)

        data_part_transformation_func = partial(array_to_input_data,
                                                idx=idx, target_array=target, task=task)

        # create labels for text data sources and remove source if there are many nans
        sources = dict((text_data_detector.new_key_name(data_part_key),
                        data_part_transformation_func(features_array=data_part, data_type=DataTypesEnum.text))
                       for (data_part_key, data_part) in data_text.items()
                       if not text_data_detector.is_full_of_nans(data_part))

        # add table features if they exist
        if table_features.size != 0:
            sources.update({'data_source_table': data_part_transformation_func
                            (features_array=table_features, data_type=DataTypesEnum.table)})

        multi_modal_data = MultiModalData(sources)

        return multi_modal_data



from copy import deepcopy
from functools import partial
from random import choice, randint, random, sample
from typing import Callable, List, Union, Tuple, TYPE_CHECKING

import numpy as np

from fedot.core.adapter import register_native
from fedot.core.dag.graph_node import GraphNode
from fedot.core.dag.graph_utils import distance_to_root_level, ordered_subnodes_hierarchy, distance_to_primary_level
from fedot.core.optimisers.advisor import RemoveType
from fedot.core.optimisers.gp_comp.gp_operators import random_graph
from fedot.core.optimisers.gp_comp.operators.operator import PopulationT, Operator
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptGraph, OptNode
from fedot.core.optimisers.opt_history_objects.individual import Individual
from fedot.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from fedot.core.optimisers.optimizer import GraphGenerationParams
from fedot.core.pipelines.pipeline_advisor import check_for_specific_operations
from fedot.core.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters


class MutationStrengthEnum(Enum):
    weak = 0.2
    mean = 1.0
    strong = 5.0


class MutationTypesEnum(Enum):
    simple = 'simple'
    growth = 'growth'
    local_growth = 'local_growth'
    reduce = 'reduce'
    single_add = 'single_add',
    single_change = 'single_change',
    single_drop = 'single_drop',
    single_edge = 'single_edge'

    none = 'none'


class Mutation(Operator):
    def __init__(self,
                 parameters: 'GPGraphOptimizerParameters',
                 requirements: PipelineComposerRequirements,
                 graph_generation_params: GraphGenerationParams):
        super().__init__(parameters, requirements)
        self.graph_generation_params = graph_generation_params

    def __call__(self, population: Union[Individual, PopulationT]) -> Union[Individual, PopulationT]:
        if isinstance(population, Individual):
            return self._mutation(population)
        return list(map(self._mutation, population))

    @staticmethod
    def get_mutation_prob(mut_id: MutationStrengthEnum, node: GraphNode,
                          default_mutation_prob: float = 0.7) -> float:
        """ Function returns mutation probability for certain node in the graph

        :param mut_id: MutationStrengthEnum mean weak or strong mutation
        :param node: root node of the graph
        :param default_mutation_prob: mutation probability used when mutation_id is invalid
        :return mutation_prob: mutation probability
        """
        if mut_id in list(MutationStrengthEnum):
            mutation_strength = mut_id.value
            mutation_prob = mutation_strength / (distance_to_primary_level(node) + 1)
        else:
            mutation_prob = default_mutation_prob
        return mutation_prob

    def _mutation(self, individual: Individual) -> Individual:
        """ Function applies mutation operator to graph """

        for _ in range(self.parameters.max_num_of_operator_attempts):
            new_graph = deepcopy(individual.graph)
            num_mut = max(int(round(np.random.lognormal(0, sigma=0.5))), 1)

            new_graph, mutation_names = self._adapt_and_apply_mutations(new_graph, num_mut)

            is_correct_graph = self.graph_generation_params.verifier(new_graph)
            if is_correct_graph:
                parent_operator = ParentOperator(type_='mutation', operators=tuple(mutation_names),
                                                 parent_individuals=individual)
                return Individual(new_graph, parent_operator)

        self.log.debug('Number of mutation attempts exceeded. '
                       'Please check composer requirements for correctness.')

        return individual

    def _adapt_and_apply_mutations(self, new_graph: OptGraph, num_mut: int) -> Tuple[OptGraph, List[str]]:
        """Apply mutation in several iterations with specific adaptation of each graph"""

        mutation_types = self.parameters.mutation_types
        is_static_mutation_type = random() < self.parameters.static_mutation_prob
        mutation_type = choice(mutation_types)
        mutation_names = []
        for _ in range(num_mut):
            # determine mutation type
            if not is_static_mutation_type:
                mutation_type = choice(mutation_types)
            is_custom_mutation = isinstance(mutation_type, Callable)

            if self._will_mutation_be_applied(mutation_type):
                # get the mutation function and adapt it
                mutation_func = self._get_mutation_func(mutation_type)
                new_graph = mutation_func(new_graph, requirements=self.requirements,
                                          params=self.graph_generation_params,
                                          opt_params=self.parameters)
                # log mutation
                mutation_names.append(str(mutation_type))
                if is_custom_mutation:
                    # custom mutation occurs once
                    break
        return new_graph, mutation_names

    def _will_mutation_be_applied(self, mutation_type: Union[MutationTypesEnum, Callable]) -> bool:
        return random() <= self.parameters.mutation_prob and mutation_type is not MutationTypesEnum.none

    @register_native
    def _simple_mutation(self, graph: OptGraph, **kwargs) -> OptGraph:
        """
        This type of mutation is passed over all nodes of the tree started from the root node and changes
        nodes’ operations with probability - 'node mutation probability'
        which is initialised inside the function

        :param graph: graph to mutate
        """

        exchange_node = self.graph_generation_params.node_factory.exchange_node
        visited_nodes = set()

        def replace_node_to_random_recursive(node: OptNode) -> OptGraph:
            if node not in visited_nodes and random() < node_mutation_probability:
                new_node = exchange_node(node)
                if new_node:
                    graph.update_node(node, new_node)
                # removed node must not be visited because it's outdated
                visited_nodes.add(node)
                # new node must not mutated if encountered further during traverse
                visited_nodes.add(new_node)
                for parent in node.nodes_from:
                    replace_node_to_random_recursive(parent)

        node_mutation_probability = self.get_mutation_prob(mut_id=self.parameters.mutation_strength,
                                                           node=graph.root_node)

        replace_node_to_random_recursive(graph.root_node)

        return graph

    @register_native
    def _single_edge_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        """
        This mutation adds new edge between two random nodes in graph.

        :param graph: graph to mutate
        """
        old_graph = deepcopy(graph)

        for _ in range(self.parameters.max_num_of_operator_attempts):
            if len(graph.nodes) < 2 or graph.depth > self.requirements.max_depth:
                return graph

            source_node, target_node = sample(graph.nodes, 2)

            nodes_not_cycling = (target_node.descriptive_id not in
                                 [n.descriptive_id for n in ordered_subnodes_hierarchy(source_node)])
            if nodes_not_cycling and (source_node not in target_node.nodes_from):
                graph.connect_nodes(source_node, target_node)
                break

        if graph.depth > self.requirements.max_depth:
            return old_graph
        return graph

    @register_native
    def _add_intermediate_node(self, graph: OptGraph, node_to_mutate: OptNode) -> OptGraph:
        # add between node and parent
        new_node = self.graph_generation_params.node_factory.get_parent_node(node_to_mutate, is_primary=False)
        if not new_node:
            return graph

        # rewire old children to new parent
        new_node.nodes_from = node_to_mutate.nodes_from
        node_to_mutate.nodes_from = [new_node]

        # add new node to graph
        graph.add_node(new_node)
        return graph

    @register_native
    def _add_separate_parent_node(self, graph: OptGraph, node_to_mutate: OptNode) -> OptGraph:
        # add as separate parent
        for iter_num in range(randint(1, 3)):
            new_node = self.graph_generation_params.node_factory.get_parent_node(node_to_mutate, is_primary=True)
            if not new_node:
                # there is no possible operators
                break
            if node_to_mutate.nodes_from:
                node_to_mutate.nodes_from.append(new_node)
            else:
                node_to_mutate.nodes_from = [new_node]
            graph.nodes.append(new_node)
        return graph

    @register_native
    def _add_as_child(self, graph: OptGraph, node_to_mutate: OptNode) -> OptGraph:
        # add as child
        old_node_children = graph.node_children(node_to_mutate)
        new_node_child = choice(old_node_children) if old_node_children else None
        operation_id = node_to_mutate.content['name']

        if check_for_specific_operations(operation_id):
            # data source, exog_ts and custom models moving is useless
            return graph

        new_node = self.graph_generation_params.node_factory.get_node(is_primary=False)
        if not new_node:
            return graph
        graph.add_node(new_node)
        graph.connect_nodes(node_parent=node_to_mutate, node_child=new_node)
        if new_node_child:
            graph.connect_nodes(node_parent=new_node, node_child=new_node_child)
            graph.disconnect_nodes(node_parent=node_to_mutate, node_child=new_node_child)

        return graph

    @register_native
    def _single_add_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        """
        Add new node between two sequential existing modes

        :param graph: graph to mutate
        """

        if graph.depth >= self.requirements.max_depth:
            # add mutation is not possible
            return graph

        node_to_mutate = choice(graph.nodes)

        single_add_strategies = [self._add_as_child, self._add_separate_parent_node]
        if node_to_mutate.nodes_from:
            single_add_strategies.append(self._add_intermediate_node)
        strategy = choice(single_add_strategies)

        result = strategy(graph, node_to_mutate)
        return result

    @register_native
    def _single_change_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        """
        Change node between two sequential existing modes.

        :param graph: graph to mutate
        """
        node = choice(graph.nodes)
        new_node = self.graph_generation_params.node_factory.exchange_node(node)
        if not new_node:
            return graph
        graph.update_node(node, new_node)
        return graph

    @register_native
    def _single_drop_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        """
        Drop single node from graph.

        :param graph: graph to mutate
        """
        node_to_del = choice(graph.nodes)
        node_name = node_to_del.content['name']
        removal_type = self.graph_generation_params.advisor.can_be_removed(node_to_del)
        # we can't delete all data source nodes
        if 'data_source' in node_name and self.only_one_data_source_node(graph):
            return graph
        if removal_type == RemoveType.with_direct_children:
            # TODO refactor workaround with data_source
            nodes_to_delete = \
                [n for n in graph.nodes if str(node_name) in n.descriptive_id and
                 n.descriptive_id.count('data_source') == 1]
            for child_node in nodes_to_delete:
                graph.delete_node(child_node)
            graph.delete_node(node_to_del)
        elif removal_type == RemoveType.with_parents:
            graph.delete_subtree(node_to_del)
        elif removal_type != RemoveType.forbidden:
            graph.delete_node(node_to_del)
            if node_to_del.nodes_from:
                children = graph.node_children(node_to_del)
                for child in children:
                    if child.nodes_from:
                        child.nodes_from.extend(node_to_del.nodes_from)
                    else:
                        child.nodes_from = node_to_del.nodes_from
        return graph

    @register_native
    def _tree_growth(self, graph: OptGraph, local_growth: bool = True) -> OptGraph:
        """
        This mutation selects a random node in a tree, generates new subtree,
        and replaces the selected node's subtree.

        :param graph: graph to mutate
        :param local_growth: if true then maximal depth of new subtree equals depth of tree located in
        selected random node, if false then previous depth of selected node doesn't affect to
        new subtree depth, maximal depth of new subtree just should satisfy depth constraint in parent tree
        """
        node_from_graph = choice(graph.nodes)
        if local_growth:
            max_depth = distance_to_primary_level(node_from_graph)
            is_primary_node_selected = (not node_from_graph.nodes_from) or (node_from_graph != graph.root_node and
                                                                            randint(0, 1))
        else:
            max_depth = self.requirements.max_depth - distance_to_root_level(graph, node_from_graph)
            is_primary_node_selected = \
                distance_to_root_level(graph, node_from_graph) >= self.requirements.max_depth and randint(0, 1)
        if is_primary_node_selected:
            new_subtree = self.graph_generation_params.node_factory.get_node(is_primary=True)
            if not new_subtree:
                return graph
        else:
            new_subtree = random_graph(self.graph_generation_params, self.requirements, max_depth).root_node
        graph.update_subtree(node_from_graph, new_subtree)
        return graph

    @register_native
    def _growth_mutation(self, graph: OptGraph, local_growth: bool = True, **kwargs) -> OptGraph:
        """
        This mutation adds new nodes to the graph (just single node between existing nodes or new subtree).

        :param graph: graph to mutate
        :param local_growth: if true then maximal depth of new subtree equals depth of tree located in
        selected random node, if false then previous depth of selected node doesn't affect to
        new subtree depth, maximal depth of new subtree just should satisfy depth constraint in parent tree
        """

        if random() > 0.5:
            # simple growth (one node can be added)
            return self._single_add_mutation(graph, self.requirements.max_depth)
        else:
            # advanced growth (several nodes can be added)
            return self._tree_growth(graph, local_growth)

    @register_native
    def _reduce_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        """
        Selects a random node in a tree, then removes its subtree. If the current arity of the node's
        parent is more than the specified minimal arity, then the selected node is also removed.
        Otherwise, it is replaced by a random primary node.

        :param graph: graph to mutate
        """
        if len(graph.nodes) == 1:
            return graph

        nodes = [node for node in graph.nodes if node is not graph.root_node]
        node_to_del = choice(nodes)
        children = graph.node_children(node_to_del)
        is_possible_to_delete = all([len(child.nodes_from) - 1 >= self.requirements.min_arity for child in children])
        if is_possible_to_delete:
            graph.delete_subtree(node_to_del)
        else:
            primary_node = self.graph_generation_params.node_factory.get_node(is_primary=True)
            if not primary_node:
                return graph
            graph.update_subtree(node_to_del, primary_node)
        return graph

    @register_native
    def _no_mutation(self, graph: OptGraph, *args, **kwargs) -> OptGraph:
        return graph

    def _get_mutation_func(self, mutation_type: Union[MutationTypesEnum, Callable]) -> Callable:
        if isinstance(mutation_type, Callable):
            mutation_func = mutation_type
        else:
            mutation_func = self.mutation_by_type(mutation_type)
        return self.graph_generation_params.adapter.adapt_func(mutation_func)

    def mutation_by_type(self, mutation_type: MutationTypesEnum) -> Callable:
        mutations = {
            MutationTypesEnum.none: self._no_mutation,
            MutationTypesEnum.simple: self._simple_mutation,
            MutationTypesEnum.growth: partial(self._growth_mutation, local_growth=False),
            MutationTypesEnum.local_growth: partial(self._growth_mutation, local_growth=True),
            MutationTypesEnum.reduce: self._reduce_mutation,
            MutationTypesEnum.single_add: self._single_add_mutation,
            MutationTypesEnum.single_edge: self._single_edge_mutation,
            MutationTypesEnum.single_drop: self._single_drop_mutation,
            MutationTypesEnum.single_change: self._single_change_mutation,
        }
        if mutation_type in mutations:
            return mutations[mutation_type]
        else:
            raise ValueError(f'Required mutation type is not found: {mutation_type}')

    def only_one_data_source_node(self, graph: OptGraph):
        count = 0
        for node in graph.nodes:
            if 'data_source' in node.content['name']:
                if count == 1:
                    return False
                count += 1
        return count == 1



import os

import numpy as np
import pandas as pd

from fedot.api.main import Fedot
from test.unit.api.test_api_cli_params import project_root_path


def load_train_test_dataframes() -> (pd.DataFrame, pd.DataFrame):
    """ Load data for multitask regression / classification problem """
    data_path = os.path.join(project_root_path, 'examples/data')
    train_df = pd.read_csv(os.path.join(data_path, 'train_synthetic_regression_classification.csv'))
    test_df = pd.read_csv(os.path.join(data_path, 'test_synthetic_regression_classification.csv'))

    # Prepare numpy arrays into dictionaries
    # For regression and classification features are the same
    mm_train_features = {'regression': np.array(train_df[['feature_1', 'feature_2']]),
                         'classification': np.array(train_df[['feature_1', 'feature_2']])}
    # Target is different for train
    mm_train_target = {'regression': np.array(train_df['concentration']),
                       'classification': np.array(train_df['class'])}
    # For test features the same
    mm_test_features = {'regression': np.array(test_df[['feature_1', 'feature_2']]),
                        'classification': np.array(test_df[['feature_1', 'feature_2']])}
    return mm_train_features, mm_train_target, mm_test_features


def launch_multitask_api_example():
    """
    Demonstration of an example with running a multitask pipeline composing procedure through FEDOT API.
    Synthetic data is used. Task: predict the category of the substance (column "class") <- classification,
    and then predict the concentration based on the predicted category (column "concentration") <- regression.
    """
    train_features, train_target, test_features = load_train_test_dataframes()

    # The priority of the task is determined by the order. So, main task is regression
    problem = 'regression/classification'

    # TODO finish this example - it is not working now and represents the desired interface for multitask
    model = Fedot(problem=problem, timeout=5)
    model.fit(features=train_features,
              target=train_target)

    predicted_concentrations, predicted_classes = model.predict(features=test_features)
    print(f'Predicted classes: {predicted_classes}')
    print(f'Predicted concentrations: {predicted_concentrations}')


if __name__ == '__main__':
    launch_multitask_api_example()



import os
import tarfile


def unpack_archived_data(archive_name: str):
    archive_path = os.path.abspath(f'{archive_name}.tar.gz')
    if (os.path.exists(archive_path) and
            os.path.basename(archive_name) not in os.listdir(os.path.dirname(archive_path))):
        with tarfile.open(archive_path) as file:
            def is_within_directory(directory, target):
                
                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)
            
                prefix = os.path.commonprefix([abs_directory, abs_target])
                
                return prefix == abs_directory
            
            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
            
                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")
            
                tar.extractall(path, members, numeric_owner=numeric_owner) 
                
            
            safe_extract(file, path=os.path.dirname(archive_path))
        print('Unpacking finished')
    else:
        print('Archive already unpacked')



from typing import Collection, Optional, Sequence, Tuple, Union

from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.composer.composer import Composer
from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptGraph
from fedot.core.optimisers.objective import PipelineObjectiveEvaluate
from fedot.core.optimisers.objective.data_source_splitter import DataSourceSplitter
from fedot.core.optimisers.optimizer import GraphOptimizer
from fedot.core.pipelines.pipeline import Pipeline


class GPComposer(Composer):
    """
    Genetic programming based composer

    :param optimizer: optimizer generated in ComposerBuilder.
    :param composer_requirements: requirements for composition process.
    :param pipelines_cache: Cache manager for fitted models, optional.
    :param preprocessing_cache: Cache manager for optional preprocessing encoders and imputers, optional.
    """

    def __init__(self, optimizer: GraphOptimizer,
                 composer_requirements: PipelineComposerRequirements,
                 pipelines_cache: Optional[OperationsCache] = None,
                 preprocessing_cache: Optional[PreprocessingCache] = None):
        super().__init__(optimizer, composer_requirements)
        self.composer_requirements = composer_requirements
        self.pipelines_cache: Optional[OperationsCache] = pipelines_cache
        self.preprocessing_cache: Optional[PreprocessingCache] = preprocessing_cache

        self.best_models: Collection[Pipeline] = ()

    def compose_pipeline(self, data: Union[InputData, MultiModalData]) -> Union[Pipeline, Sequence[Pipeline]]:
        # Define data source
        data_producer = DataSourceSplitter(self.composer_requirements.cv_folds,
                                           self.composer_requirements.validation_blocks,
                                           shuffle=True).build(data)
        # Define objective function
        n_jobs_for_evaluation = 1
        # TODO implement dispatcher selection
        objective_evaluator = PipelineObjectiveEvaluate(self.optimizer.objective, data_producer,
                                                        self.composer_requirements.max_pipeline_fit_time,
                                                        self.composer_requirements.validation_blocks,
                                                        self.pipelines_cache, self.preprocessing_cache,
                                                        eval_n_jobs=n_jobs_for_evaluation)
        objective_function = objective_evaluator.evaluate

        # Define callback for computing intermediate metrics if needed
        if self.composer_requirements.collect_intermediate_metric:
            self.optimizer.set_evaluation_callback(objective_evaluator.evaluate_intermediate_metrics)

        # Finally, run optimization process
        opt_result = self.optimizer.optimise(objective_function)

        best_model, self.best_models = self._convert_opt_results_to_pipeline(opt_result)
        self.log.info('GP composition finished')
        return best_model

    def _convert_opt_results_to_pipeline(self, opt_result: Sequence[OptGraph]) -> Tuple[Pipeline, Sequence[Pipeline]]:
        adapter = self.optimizer.graph_generation_params.adapter
        multi_objective = self.optimizer.objective.is_multi_objective
        best_pipelines = [adapter.restore(graph) for graph in opt_result]
        chosen_best_pipeline = best_pipelines if multi_objective else best_pipelines[0]
        return chosen_best_pipeline, best_pipelines

    @staticmethod
    def tune_pipeline(pipeline: Pipeline, data: InputData, time_limit):
        raise NotImplementedError()



from copy import deepcopy
from datetime import timedelta
from os import PathLike
from typing import List, Optional, Tuple, Union, Sequence

import func_timeout

from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.dag.graph import Graph
from fedot.core.dag.graph_delegate import GraphDelegate
from fedot.core.dag.graph_node import GraphNode
from fedot.core.dag.graph_utils import distance_to_primary_level
from fedot.core.dag.linked_graph import LinkedGraph
from fedot.core.data.data import InputData, OutputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import default_log
from fedot.core.operations.data_operation import DataOperation
from fedot.core.operations.model import Model
from fedot.core.optimisers.timer import Timer
from fedot.core.pipelines.node import Node, PrimaryNode, SecondaryNode
from fedot.core.pipelines.template import PipelineTemplate
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.utilities.serializable import Serializable
from fedot.core.utils import copy_doc
from fedot.core.visualisation.graph_viz import NodeColorType
from fedot.core.visualisation.pipeline_specific_visuals import PipelineVisualizer
from fedot.preprocessing.preprocessing import DataPreprocessor, update_indices_for_time_series

ERROR_PREFIX = 'Invalid pipeline configuration:'


class Pipeline(GraphDelegate, Serializable):
    """Base class used for composite model structure definition

    Args:
        nodes: :obj:`Node` object(s)
    """

    def __init__(self, nodes: Union[Node, Sequence[Node]] = ()):
        super().__init__(nodes, _graph_nodes_to_pipeline_nodes)

        self.computation_time = None
        self.log = default_log(self)

        # Define data preprocessor
        self.preprocessor = DataPreprocessor()

    def fit_from_scratch(self, input_data: Union[InputData, MultiModalData] = None):
        """[Obsolete] Method used for training the pipeline without using saved information

        Args:
            input_data: data used for operation training
        """

        # Clean all saved states and fit all operations
        self.unfit()
        self.fit(input_data)

    def _fit_with_time_limit(self, input_data: Optional[InputData],
                             time: timedelta) -> OutputData:
        """Runs training process in all of the pipeline nodes starting with root with time limit.

        Todo:
            unresolved sentence

        Args:
            input_data: data used for operations training
            use_fitted_operations: flag defining whether to use saved information about previous executions or not
            time: time constraint for operations fitting process (in minutes)

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        time = int(time.total_seconds())
        process_state_dict = {}
        fitted_operations = []
        try:
            func_timeout.func_timeout(
                time, self._fit,
                args=(input_data, process_state_dict, fitted_operations)
            )
        except func_timeout.FunctionTimedOut:
            raise TimeoutError(f'Pipeline fitness evaluation time limit is expired (more then {time} seconds)')

        self.computation_time = process_state_dict['computation_time_in_seconds']
        for node_num, _ in enumerate(self.nodes):
            self.nodes[node_num].fitted_operation = fitted_operations[node_num]
        return process_state_dict['train_predicted']

    def _fit(self, input_data: Optional[InputData] = None,
             process_state_dict: dict = None, fitted_operations: list = None) -> Optional[OutputData]:
        """Runs training process in all of the pipeline nodes starting with root

        Args:
            input_data: data used for operation training
            use_fitted_operations: flag defining whether to use saved information about previous executions or not
            process_state_dict: dictionary used for saving required pipeline parameters
                (which were changed inside the process) in case of operations fit time control (when process created)
            fitted_operations: list used for saving fitted operations of pipeline nodes

        Returns:
            Optional[OutputData]: values predicted on the provided ``input_data`` or nothing
            in case of the time controlled call
        """

        with Timer() as t:
            computation_time_update = not self.root_node.fitted_operation or self.computation_time is None
            train_predicted = self.root_node.fit(input_data=input_data)
            if computation_time_update:
                self.computation_time = round(t.minutes_from_start, 3)

        if process_state_dict is None:
            return train_predicted
        else:
            process_state_dict['train_predicted'] = train_predicted
            process_state_dict['computation_time_in_seconds'] = self.computation_time
            for node in self.nodes:
                fitted_operations.append(node.fitted_operation)

    def fit(self, input_data: Union[InputData, MultiModalData],
            time_constraint: Optional[timedelta] = None, n_jobs: int = 1) -> OutputData:
        """
        Runs training process in all the pipeline nodes starting with root

        Args:
            input_data: data used for operations training
            time_constraint: time constraint for operations fitting (in seconds)
            n_jobs: number of threads for nodes fitting

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """
        self.replace_n_jobs_in_nodes(n_jobs)

        copied_input_data = deepcopy(input_data)
        copied_input_data = self.preprocessor.obligatory_prepare_for_fit(copied_input_data)
        # Make additional preprocessing if it is needed
        copied_input_data = self.preprocessor.optional_prepare_for_fit(pipeline=self,
                                                                       data=copied_input_data)
        copied_input_data = self.preprocessor.convert_indexes_for_fit(pipeline=self,
                                                                      data=copied_input_data)
        copied_input_data = self._assign_data_to_nodes(copied_input_data)
        if time_constraint is None:
            train_predicted = self._fit(input_data=copied_input_data)
        else:
            train_predicted = self._fit_with_time_limit(input_data=copied_input_data, time=time_constraint)
        return train_predicted

    @property
    def is_fitted(self) -> bool:
        """Property showing whether pipeline is fitted

        Returns:
            flag showing if all of the pipeline nodes are fitted already
        """

        return all(node.fitted_operation is not None for node in self.nodes)

    def unfit(self, mode='all', unfit_preprocessor: bool = True):
        """Removes fitted operations for chosen type of nodes.

        Args:
            mode: the name of mode

                .. details:: possible ``mode`` options:

                        - ``all`` -> (default) All models will be unfitted
                        - ``data_operations`` -> All data operations will be unfitted

            unfit_preprocessor: should we unfit preprocessor
        """

        for node in self.nodes:
            if mode == 'all' or (mode == 'data_operations' and isinstance(node.content['name'], DataOperation)):
                node.unfit()

        if unfit_preprocessor:
            self.unfit_preprocessor()

    def unfit_preprocessor(self):
        self.preprocessor = DataPreprocessor()

    def try_load_from_cache(self, cache: Optional[OperationsCache], preprocessing_cache: Optional[PreprocessingCache],
                            fold_id: Optional[int] = None):
        """
        Tries to load pipeline nodes if ``cache`` is provided

        Args:
            cache: pipeline nodes cacher
            fold_id: optional part of the cache item UID
               (can be used to specify the number of CV fold)

        Returns:
            bool: indicating if at least one node was loaded
        """
        if cache is not None:
            cache.try_load_into_pipeline(self, fold_id)
        if preprocessing_cache is not None:
            preprocessing_cache.try_load_preprocessor(self, fold_id)

    def predict(self, input_data: Union[InputData, MultiModalData], output_mode: str = 'default') -> OutputData:
        """Runs the predict process in all of the pipeline nodes starting with root

        input_data: data for prediction
        output_mode: desired form of output for operations

            .. details:: possible ``output_mode`` options:

                - ``default`` -> (as is, default)
                - ``labels`` -> (numbers of classes - for classification)
                - ``probs`` -> (probabilities - for classification == default)
                - ``full_probs`` -> (return all probabilities - for binary classification)

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        if not self.is_fitted:
            ex = 'Pipeline is not fitted yet'
            self.log.error(ex)
            raise ValueError(ex)

        # Make copy of the input data to avoid performing inplace operations
        copied_input_data = deepcopy(input_data)
        copied_input_data = self.preprocessor.obligatory_prepare_for_predict(copied_input_data)
        # Make additional preprocessing if it is needed
        copied_input_data = self.preprocessor.optional_prepare_for_predict(pipeline=self,
                                                                           data=copied_input_data)
        copied_input_data = self.preprocessor.convert_indexes_for_predict(pipeline=self,
                                                                          data=copied_input_data)
        copied_input_data = update_indices_for_time_series(copied_input_data)

        copied_input_data = self._assign_data_to_nodes(copied_input_data)

        result = self.root_node.predict(input_data=copied_input_data, output_mode=output_mode)

        result = self.preprocessor.restore_index(copied_input_data, result)
        # Prediction should be converted into source labels (if it is needed)
        if output_mode == 'labels':
            result.predict = self.preprocessor.apply_inverse_target_encoding(result.predict)
        return result

    def save(self, path: str = None, create_subdir: bool = True, is_datetime_in_path: bool = False) -> Tuple[str, dict]:
        """
        Saves the pipeline to JSON representation with pickled fitted operations

        Args:
            path: custom path to dir where to save JSON or name of json file where to save pipeline.
            If only file name is specified, than absolute path to this file will be created.
            create_subdir: if True -- create one more dir in the last dir
                           if False -- save to the last dir in specified path
            is_datetime_in_path: is it required to add the datetime timestamp to the path

        Returns:
            Tuple[str, dict]: :obj:`JSON representation of the pipeline structure`,
            :obj:`dict of paths to fitted models`
        """

        template = PipelineTemplate(self)
        json_object, dict_fitted_operations = template.export_pipeline(path, root_node=self.root_node,
                                                                       create_subdir=create_subdir,
                                                                       is_datetime_in_path=is_datetime_in_path)
        return json_object, dict_fitted_operations

    def load(self, source: Union[str, dict], dict_fitted_operations: Optional[dict] = None):
        """Loads the pipeline ``JSON`` representation with pickled fitted operations.

        Args:
            source: where to load the pipeline from
            dict_fitted_operations: dictionary of the fitted operations
        """

        self.nodes = []
        template = PipelineTemplate(self)
        template.import_pipeline(source, dict_fitted_operations)
        return self

    @property
    def root_node(self) -> Optional[Node]:
        """Finds pipelines sink-node

        Returns:
            the final predictor-node
        """
        if not self.nodes:
            return None
        root = [node for node in self.nodes
                if not any(self.node_children(node))]
        if len(root) > 1:
            raise ValueError(f'{ERROR_PREFIX} More than 1 root_nodes in pipeline')
        return root[0]

    def pipeline_for_side_task(self, task_type: TaskTypesEnum) -> 'Pipeline':
        """Returns pipeline formed from the last node solving the given problem and all its parents

        Args:
            task_type: task type of the last node to search for

        Returns:
            pipeline formed from the last node solving the given problem and all of its parents
        """

        max_distance = 0
        side_root_node = None
        for node in self.nodes:
            if (task_type in node.operation.acceptable_task_types and
                    isinstance(node.operation, Model) and
                    distance_to_primary_level(node) >= max_distance):
                side_root_node = node
                max_distance = distance_to_primary_level(node)

        pipeline = Pipeline(side_root_node)
        pipeline.preprocessor = self.preprocessor
        return pipeline

    def _assign_data_to_nodes(self, input_data: Union[InputData, MultiModalData]) -> Optional[InputData]:
        """In case of provided ``input_data`` is of type :class:`MultiModalData`
        assigns :attr:`PrimaryNode.node_data` from the ``input_data``

        Args:
            input_data: data to assign to :attr:`PrimaryNode.node_data`

        Returns:
            ``None`` in case of :class:`MultiModalData` and ``input_data`` otherwise
        """

        if isinstance(input_data, MultiModalData):
            for node in (n for n in self.nodes if isinstance(n, PrimaryNode)):
                if node.operation.operation_type in input_data:
                    node.node_data = input_data[node.operation.operation_type]
                    node.direct_set = True
                else:
                    raise ValueError(f'No data for primary node {node}')
            return None
        return input_data

    @property
    def structure(self) -> str:
        """ Structural information about the pipeline

            Returns:
                string with pipeline structure
        """
        return '\n'.join([str(self), *(f'{node.operation.operation_type} - {node.parameters}' for node in self.nodes)])

    def print_structure(self):
        """ Prints structure of the pipeline
        """

        print(
            'Pipeline structure:',
            self.structure,
            sep='\n'
        )

    def replace_n_jobs_in_nodes(self, n_jobs: int):
        """
        Changes number of jobs for nodes

        :param n_jobs: required number of the jobs to assign to the nodes
        """
        for node in self.nodes:
            for param in ['n_jobs', 'num_threads']:
                if param in node.content['params']:
                    node.content['params'][param] = n_jobs
                    # workaround for lgbm paramaters
                    if node.content['name'] == 'lgbm':
                        node.content['params']['num_threads'] = n_jobs
                        node.content['params']['n_jobs'] = n_jobs

    @copy_doc(Graph)
    def show(self, save_path: Optional[Union[PathLike, str]] = None, engine: Optional[str] = None,
             node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
             node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
             edge_curvature_scale: Optional[float] = None):
        PipelineVisualizer(self).visualise(save_path, engine, node_color, dpi, node_size_scale, font_size_scale)


def _graph_nodes_to_pipeline_nodes(operator: LinkedGraph, nodes: Sequence[Node]):
    """
    Method to update nodes type after performing some action on the pipeline
        via GraphOperator, if any of them are of GraphNode type

    Args:
        nodes: :obj:`Node` object(s)
    """

    for node in nodes:
        if not isinstance(node, GraphNode):
            continue
        if node.nodes_from and not isinstance(node, SecondaryNode):
            operator.update_node(old_node=node,
                                 new_node=SecondaryNode(nodes_from=node.nodes_from,
                                                        content=node.content))
        # TODO: avoid internal access use operator.delete_node
        elif not node.nodes_from and not operator.node_children(node) and node != operator.root_node:
            operator.nodes.remove(node)
        elif not node.nodes_from and not isinstance(node, PrimaryNode):
            operator.update_node(old_node=node,
                                 new_node=PrimaryNode(nodes_from=node.nodes_from,
                                                      content=node.content))



from abc import ABC, abstractmethod
from copy import deepcopy
from datetime import timedelta
from typing import Callable, ClassVar

import numpy as np
from hyperopt.early_stop import no_progress_loss

from fedot.core.log import default_log
from fedot.core.optimisers.objective import ObjectiveEvaluate
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.search_space import SearchSpace

MAX_METRIC_VALUE = np.inf


class HyperoptTuner(ABC):
    """
    Base class for hyperparameters optimization based on hyperopt library
    
    Args:
      iterations: max number of iterations
      search_space: SearchSpace instance
      algo: algorithm for hyperparameters optimization with signature similar to :obj:`hyperopt.tse.suggest`
    """

    def __init__(self, objective_evaluate: ObjectiveEvaluate,
                 iterations=100, early_stopping_rounds=None,
                 timeout: timedelta = timedelta(minutes=5),
                 search_space: ClassVar = SearchSpace(),
                 algo: Callable = None,
                 n_jobs: int = -1):
        self.iterations = iterations
        iteration_stop_count = early_stopping_rounds or max(100, int(np.sqrt(iterations) * 10))
        self.early_stop_fn = no_progress_loss(iteration_stop_count=iteration_stop_count)
        self.max_seconds = int(timeout.seconds) if timeout is not None else None
        self.init_pipeline = None
        self.init_metric = None
        self.obtained_metric = None
        self.objective_evaluate = objective_evaluate
        self._default_metric_value = MAX_METRIC_VALUE
        self.search_space = search_space
        self.algo = algo
        self.n_jobs = n_jobs

        self.log = default_log(self)

    @abstractmethod
    def tune(self, pipeline: Pipeline) -> Pipeline:
        """
        Function for hyperparameters tuning on the pipeline

        Args:
          pipeline: Pipeline for which hyperparameters tuning is needed

        Returns:
          graph with optimized hyperparameters
        """
        raise NotImplementedError()

    def get_metric_value(self, pipeline: Pipeline) -> float:
        """
        Method calculates metric for algorithm validation
        
        Args:
          pipeline: Pipeline to evaluate

        Returns:
          value of loss function
        """
        pipeline.unfit()
        pipeline_fitness = self.objective_evaluate.evaluate(pipeline)
        metric_value = pipeline_fitness.value
        if not pipeline_fitness.valid:
            return self._default_metric_value
        return metric_value

    def init_check(self, pipeline: Pipeline) -> None:
        """
        Method get metric on validation set before start optimization

        Args:
          pipeline: Pipeline to calculate objective
        """
        self.log.info('Hyperparameters optimization start')

        # Train pipeline
        self.init_pipeline = deepcopy(pipeline)

        self.init_metric = self.get_metric_value(pipeline=self.init_pipeline)
        self.log.message(f'Initial pipeline: {self.init_pipeline.structure} \n'
                         f'Initial metric: {abs(self.init_metric):.3f}')

    def final_check(self, tuned_pipeline: Pipeline):
        """
        Method propose final quality check after optimization process

        Args:
          tuned_pipeline: Tuned pipeline to calculate objective
        """

        self.obtained_metric = self.get_metric_value(pipeline=tuned_pipeline)

        if self.obtained_metric == self._default_metric_value:
            self.obtained_metric = None

        self.log.info('Hyperparameters optimization finished')

        prefix_tuned_phrase = 'Return tuned pipeline due to the fact that obtained metric'
        prefix_init_phrase = 'Return init pipeline due to the fact that obtained metric'

        # 5% deviation is acceptable
        deviation = (self.init_metric / 100.0) * 5
        init_metric = self.init_metric + deviation * np.sign(self.init_metric)
        if self.obtained_metric is None:
            self.log.info(f'{prefix_init_phrase} is None. Initial metric is {abs(init_metric):.3f}')
            final_pipeline = self.init_pipeline

        elif self.obtained_metric <= init_metric:
            self.log.info(f'{prefix_tuned_phrase} {abs(self.obtained_metric):.3f} equal or '
                          f'better than initial (+ 5% deviation) {abs(init_metric):.3f}')
            final_pipeline = tuned_pipeline
        else:
            self.log.info(f'{prefix_init_phrase} {abs(self.obtained_metric):.3f} '
                          f'worse than initial (+ 5% deviation) {abs(init_metric):.3f}')
            final_pipeline = self.init_pipeline
        self.log.message(f'Final pipeline: {final_pipeline.structure}')
        if self.obtained_metric is not None:
            self.log.message(f'Final metric: {abs(self.obtained_metric):.3f}')
        else:
            self.log.message(f'Final metric is None')
        return final_pipeline



from copy import deepcopy
from typing import Tuple, Union

from sklearn.model_selection import train_test_split

from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum


def _split_time_series(data: InputData, task, *args, **kwargs):
    """ Split time series data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    """

    input_features = data.features
    input_target = data.target
    forecast_length = task.task_params.forecast_length

    if kwargs.get('validation_blocks') is not None:
        # It is required to split data for in-sample forecasting
        forecast_length = forecast_length * kwargs.get('validation_blocks')
        x_train = input_features[:-forecast_length]
        x_test = input_features

        y_train = input_target[:-forecast_length]
        y_test = input_target[-forecast_length:]
    else:
        # Source time series divide into two parts
        x_train = input_features[:-forecast_length]
        x_test = input_features[:-forecast_length]

        y_train = input_target[:-forecast_length]
        y_test = input_target[-forecast_length:]

    idx_train = data.idx[:-forecast_length]
    idx_test = data.idx[-forecast_length:]

    # Prepare data to train the operation
    train_data = InputData(idx=idx_train, features=x_train, target=y_train,
                           task=task, data_type=DataTypesEnum.ts,
                           supplementary_data=data.supplementary_data)

    test_data = InputData(idx=idx_test, features=x_test, target=y_test,
                          task=task, data_type=DataTypesEnum.ts,
                          supplementary_data=data.supplementary_data)
    return train_data, test_data


def _split_multi_time_series(data: InputData, task, *args, **kwargs):
    """ Split multi_ts time series data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    """

    input_features = data.features
    input_target = data.target
    forecast_length = task.task_params.forecast_length

    if kwargs.get('validation_blocks') is not None:
        # It is required to split data for in-sample forecasting
        forecast_length = forecast_length * kwargs.get('validation_blocks')
        x_train = input_features[:-forecast_length]
        x_test = input_features

        y_train = input_target[:-forecast_length]
        y_test = input_target[-forecast_length:, 0]

    else:
        # Source time series divide into two parts
        x_train = input_features[:-forecast_length]
        x_test = input_features[:-forecast_length]

        y_train = input_target[:-forecast_length]
        y_test = input_target[-forecast_length:, 0]

    idx_train = data.idx[:-forecast_length]
    idx_test = data.idx[-forecast_length:]

    # Prepare data to train the operation
    train_data = InputData(idx=idx_train, features=x_train, target=y_train,
                           task=task, data_type=DataTypesEnum.multi_ts,
                           supplementary_data=data.supplementary_data)

    test_data = InputData(idx=idx_test, features=x_test, target=y_test,
                          task=task, data_type=DataTypesEnum.multi_ts,
                          supplementary_data=data.supplementary_data)

    return train_data, test_data


def _split_any(data: InputData, task, data_type, split_ratio, with_shuffle=False, **kwargs):
    """ Split any data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    :param split_ratio: threshold for partitioning
    :param data_type type of data to split
    :param with_shuffle: is data needed to be shuffled or not
    """

    if not 0. < split_ratio < 1.:
        raise ValueError('Split ratio must belong to the interval (0; 1)')
    random_state = 42

    # Predictors and target
    input_features = data.features
    input_target = data.target
    idx = data.idx
    idx_train, idx_test, x_train, x_test, y_train, y_test = \
        train_test_split(idx,
                         input_features,
                         input_target,
                         test_size=1. - split_ratio,
                         shuffle=with_shuffle,
                         random_state=random_state)

    # Prepare data to train the operation
    train_data = InputData(idx=idx_train, features=x_train,  target=y_train,
                           task=task, data_type=data_type,
                           supplementary_data=data.supplementary_data)

    test_data = InputData(idx=idx_test, features=x_test, target=y_test,
                          task=task, data_type=data_type,
                          supplementary_data=data.supplementary_data)

    return train_data, test_data


def _split_table(data: InputData, task, split_ratio, with_shuffle=False, **kwargs):
    """ Split table data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    :param split_ratio: threshold for partitioning
    :param with_shuffle: is data needed to be shuffled or not
    """
    return _split_any(data, task, DataTypesEnum.table, split_ratio, with_shuffle)


def _split_image(data: InputData, task, split_ratio, with_shuffle=False, **kwargs):
    """ Split image data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    :param split_ratio: threshold for partitioning
    :param with_shuffle: is data needed to be shuffled or not
    """

    return _split_any(data, task, DataTypesEnum.image, split_ratio, with_shuffle)


def _split_text(data: InputData, task, split_ratio, with_shuffle=False, **kwargs):
    """ Split text data into train and test parts

    :param data: InputData object to split
    :param task: task to solve
    :param split_ratio: threshold for partitioning
    :param with_shuffle: is data needed to be shuffled or not
    """

    return _split_any(data, task, DataTypesEnum.text, split_ratio, with_shuffle)


def _train_test_single_data_setup(data: InputData, split_ratio=0.8,
                                  shuffle_flag=False, **kwargs) -> Tuple[InputData, InputData]:
    """ Function for train and test split

    :param data: InputData for train and test splitting
    :param split_ratio: threshold for partitioning
    :param shuffle_flag: is data needed to be shuffled or not

    :return train_data: InputData for train
    :return test_data: InputData for validation
    """
    # Split into train and test
    if data is not None:
        task = data.task

        split_func_dict = {
            DataTypesEnum.multi_ts: _split_multi_time_series,
            DataTypesEnum.ts: _split_time_series,
            DataTypesEnum.table: _split_table,
            DataTypesEnum.image: _split_image,
            DataTypesEnum.text: _split_text
        }

        split_func = split_func_dict.get(data.data_type, _split_table)

        train_data, test_data = split_func(data, task, split_ratio,
                                           with_shuffle=shuffle_flag,
                                           **kwargs)
    else:
        raise ValueError('InputData must be not empty')

    # Store additional information
    train_data.supplementary_data = deepcopy(data.supplementary_data)
    test_data.supplementary_data = deepcopy(data.supplementary_data)
    return train_data, test_data


def _train_test_multi_modal_data_setup(data: MultiModalData, split_ratio=0.8,
                                       shuffle_flag=False, **kwargs) -> Tuple[MultiModalData, MultiModalData]:
    train_data = MultiModalData()
    test_data = MultiModalData()
    for node in data.keys():
        data_part = data[node]
        train_data_part, test_data_part = train_test_data_setup(data_part, split_ratio, shuffle_flag, **kwargs)
        train_data[node] = train_data_part
        test_data[node] = test_data_part

    return train_data, test_data


def train_test_data_setup(data: Union[InputData, MultiModalData], split_ratio=0.8,
                          shuffle_flag=False, **kwargs) -> Tuple[Union[InputData, MultiModalData],
                                                                 Union[InputData, MultiModalData]]:
    """ Function for train and test split for both InputData and MultiModalData

    Args:
        data: data for train and test splitting
        split_ratio: threshold for partitioning
        shuffle_flag: is data needed to be shuffled or not
        kwargs: additional optional parameters such as number of validation blocks

    Returns:
        data for train, data for validation
    """
    if isinstance(data, InputData):
        train_data, test_data = _train_test_single_data_setup(data, split_ratio, shuffle_flag, **kwargs)
    elif isinstance(data, MultiModalData):
        train_data, test_data = _train_test_multi_modal_data_setup(data, split_ratio, shuffle_flag, **kwargs)
    else:
        raise ValueError(f'Dataset {type(data)} is not supported')

    return train_data, test_data



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fedot.core.log import default_log
from fedot.core.visualisation.opt_history.arg_constraint_wrapper import ArgConstraintWrapper

if TYPE_CHECKING:
    from fedot.core.visualisation.opt_viz import OptHistoryVisualizer


class HistoryVisualization(metaclass=ArgConstraintWrapper):
    """ Base class for creating visualizations of FEDOT optimization history.
    The only necessary method is 'visualize' - it must show or save the plot in any form after the call.

    One should refer the OptHistory instance as `self.history` to be able to connect one's visualization
    to `OptHistory.show()`. See the examples of connecting visualizations in the module `opt_viz.py`.

    It is good practice to be aware of constraints on your visualizations. You can either implement
    default constraints that will catch your kwarg across all the visualizations or define your single
    class specific constraints by assigning them to `constraint_checkers` class attribute.
    See `fedot.core.visualisation.opt_history.arg_constraint_wrapper.py` for examples.
    """
    constraint_checkers = []  # Use this for class-specific constraint checkers.

    def __init__(self, visualizer: OptHistoryVisualizer):
        self.visualizer = visualizer
        self.history = visualizer.history
        self.log = default_log(self)

    @abstractmethod
    def visualize(self, *args, **kwargs):
        raise NotImplementedError()

    def get_predefined_value(self, param: str):
        return self.visualizer.visuals_params.get(param)



import random
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.decomposition import KernelPCA, PCA, FastICA
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import MinMaxScaler, PolynomialFeatures, StandardScaler

from fedot.core.data.data import InputData, data_type_is_table, OutputData
from fedot.core.data.data_preprocessing import replace_inf_with_nans, convert_into_column, \
    divide_data_categorical_numerical, find_categorical_columns, data_has_categorical_features
from fedot.core.operations.evaluation.operation_implementations. \
    implementation_interfaces import DataOperationImplementation, EncodedInvariantImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class ComponentAnalysisImplementation(DataOperationImplementation):
    """ Class for applying PCA and kernel PCA models from sklearn

    Args:
        params: OpearationParameters with the arguments
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.pca = None
        self.number_of_features = None
        self.number_of_samples = None

    def fit(self, input_data: InputData):
        """The method trains the PCA model

        Args:
            input_data: data with features, target and ids for PCA training

        Returns:
            trained PCA model (optional output)
        """

        self.number_of_samples, self.number_of_features = np.array(input_data.features).shape

        if self.number_of_features > 1:
            self.check_and_correct_params()
            self.pca.fit(input_data.features)

        return self.pca

    def transform(self, input_data: InputData) -> OutputData:
        """Method for transformation tabular data using PCA

        Args:
            input_data: data with features, target and ids for PCA applying

        Returns:
            data with transformed features attribute
        """

        if self.number_of_features > 1:
            transformed_features = self.pca.transform(input_data.features)
        else:
            transformed_features = input_data.features

        # Update features
        output_data = self._convert_to_output(input_data,
                                              transformed_features)
        self.update_column_types(output_data)
        return output_data

    def check_and_correct_params(self):
        """Method check if number of features in data enough for ``n_components``
        parameter in PCA or not. And if not enough - fixes it
        """
        n_components = self.params.get('n_components')
        if type(n_components) == int:
            if n_components > self.number_of_features:
                self.params.update(n_components=self.number_of_features)
        elif n_components == 'mle':
            # Check that n_samples correctly map with n_features
            if self.number_of_samples < self.number_of_features:
                self.params.update(n_components=0.5)

        self.pca.set_params(**self.params.to_dict())

    @staticmethod
    def update_column_types(output_data: OutputData) -> OutputData:
        """Update column types after applying PCA operations
        """

        n_rows, n_cols = output_data.predict.shape
        output_data.supplementary_data.column_types['features'] = [str(float) * n_cols]
        return output_data


class PCAImplementation(ComponentAnalysisImplementation):
    """Class for applying PCA from sklearn

    Args:
        params: OperationParameters with the hyperparameters
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        if not self.params:
            # Default parameters
            default_params = {'svd_solver': 'full', 'n_components': 'mle'}
            self.params.update(**default_params)
        self.pca = PCA(**self.params.to_dict())
        self.number_of_features = None


class KernelPCAImplementation(ComponentAnalysisImplementation):
    """ Class for applying kernel PCA from sklearn

    Args:
        params: OperationParameters with the hyperparameters
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.pca = KernelPCA(**self.params.to_dict())


class FastICAImplementation(ComponentAnalysisImplementation):
    """ Class for applying FastICA from sklearn

    Args:
        params: OperationParameters with the hyperparameters
    """
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.pca = FastICA(**self.params.to_dict())


class PolyFeaturesImplementation(EncodedInvariantImplementation):
    """ Class for application of :obj:`PolynomialFeatures` operation on data,
    where only not encoded features (were not converted from categorical using
    ``OneHot encoding``) are used

    Args:
        params: OperationParameters with the arguments
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.th_columns = 10
        if not self.params:
            # Default parameters
            self.operation = PolynomialFeatures(include_bias=False)
        else:
            # Checking the appropriate params are using or not
            poly_params = {k: self.params.get(k) for k in
                           ['degree', 'interaction_only']}
            self.operation = PolynomialFeatures(include_bias=False,
                                                **poly_params)
        self.columns_to_take = None

    def fit(self, input_data: InputData):
        """ Method for fit Poly features operation """
        # Check the number of columns in source dataset
        n_rows, n_cols = input_data.features.shape
        if n_cols > self.th_columns:
            # Randomly choose subsample of features columns - 10 features
            column_indices = np.arange(n_cols)
            self.columns_to_take = random.sample(list(column_indices), self.th_columns)
            input_data = input_data.subset_features(self.columns_to_take)

        return super().fit(input_data)

    def transform(self, input_data: InputData) -> OutputData:
        """Firstly perform filtration of columns
        """

        clipped_input_data = input_data
        if self.columns_to_take is not None:
            clipped_input_data = input_data.subset_features(self.columns_to_take)
        output_data = super().transform(clipped_input_data)

        if self.columns_to_take is not None:
            # Get generated features from poly function
            generated_features = output_data.predict[:, self.th_columns:]
            # Concat source features with generated one
            all_features = np.hstack((input_data.features, generated_features))
            output_data.predict = all_features
        return output_data

    def _update_column_types(self, source_features_shape, output_data: OutputData):
        """Update column types after applying operations. If new columns added, new type for them are defined
        """

        if len(source_features_shape) < 2:
            return output_data
        else:
            cols_number_added = output_data.predict.shape[1] - source_features_shape[1]
            if cols_number_added > 0:
                # There are new columns in the table
                col_types = output_data.supplementary_data.column_types['features']
                col_types.extend([str(float)] * cols_number_added)
                output_data.supplementary_data.column_types['features'] = col_types


class ScalingImplementation(EncodedInvariantImplementation):
    """Class for application of ``Scaling operation`` on data,
    where only not encoded features (were not converted from categorical using
    ``OneHot encoding``) are used

    Args:
        params: OperationParameters with the arguments
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.operation = StandardScaler(**self.params.to_dict())


class NormalizationImplementation(EncodedInvariantImplementation):
    """Class for application of ``MinMax normalization`` operation on data,
    where only not encoded features (were not converted from categorical using
    ``OneHot encoding``) are used

    Args:
        params: OperationParameters with the arguments
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.operation = MinMaxScaler(**self.params.to_dict())


class ImputationImplementation(DataOperationImplementation):
    """Class for applying imputation on tabular data

    Args:
        params: OperationParameters with the arguments
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        default_params_categorical = {'strategy': 'most_frequent'}
        self.params_cat = {**self.params.to_dict(), **default_params_categorical}
        self.params_num = self.params.to_dict()
        self.categorical_ids = None
        self.non_categorical_ids = None
        self.ids_binary_integer_features = {}

        self.imputer_cat = SimpleImputer(**self.params_cat)
        self.imputer_num = SimpleImputer(**self.params_num)

    def fit(self, input_data: InputData):
        """The method trains ``SimpleImputer``

        Args:
            input_data: data with features
        """

        replace_inf_with_nans(input_data)

        if data_type_is_table(input_data):
            # Tabular data contains categorical features
            categorical_ids, non_categorical_ids = find_categorical_columns(input_data.features)
            numerical, categorical = divide_data_categorical_numerical(input_data, categorical_ids,
                                                                       non_categorical_ids)

            if categorical is not None and categorical.features.size > 0:
                categorical.features = convert_into_column(categorical.features)
                # Imputing for categorical values
                self.imputer_cat.fit(categorical.features)

            if numerical is not None and numerical.features.size > 0:
                numerical.features = convert_into_column(numerical.features)
                # Imputing for numerical values
                self.imputer_num.fit(numerical.features)
        else:
            # Time series or other type of non-tabular data
            input_data.features = convert_into_column(input_data.features)
            self.imputer_num.fit(input_data.features)

    def transform(self, input_data: InputData) -> OutputData:
        """Method for transformation tabular data using ``SimpleImputer``

        Args:
            input_data: data with features

        Returns:
            data with transformed features attribute
        """

        replace_inf_with_nans(input_data)

        if data_type_is_table(input_data) and data_has_categorical_features(input_data):
            features_types = input_data.supplementary_data.column_types.get('features')
            self.categorical_ids, self.non_categorical_ids = find_categorical_columns(input_data.features,
                                                                                      features_types)
            numerical, categorical = divide_data_categorical_numerical(input_data, self.categorical_ids,
                                                                       self.non_categorical_ids)

            if categorical is not None:
                categorical_features = convert_into_column(categorical.features)
                categorical_features = self.imputer_cat.transform(categorical_features)

            if numerical is not None:
                numerical_features = convert_into_column(numerical.features)

                # Features with only two unique values must be filled in a specific way
                self._find_binary_features(numerical_features)
                numerical_features = self.imputer_num.transform(numerical_features)
                numerical_features = self._correct_binary_ids_features(numerical_features)

            if categorical is not None and numerical is not None:
                # Stack both categorical and numerical features
                transformed_features = self._categorical_numerical_union(categorical_features,
                                                                         numerical_features)
            elif categorical is not None and numerical is None:
                # Dataset contain only categorical features
                transformed_features = categorical_features
            elif categorical is None and numerical is not None:
                # Dataset contain only numerical features
                transformed_features = numerical_features

        else:
            input_data.features = convert_into_column(input_data.features)
            transformed_features = self.imputer_num.transform(input_data.features)

        output_data = self._convert_to_output(input_data, transformed_features, data_type=input_data.data_type)
        return output_data

    def fit_transform(self, input_data: InputData) -> OutputData:
        """Method for training and transformation tabular data using ``SimpleImputer``

        Args:
            input_data: data with features

        Returns:
            data with transformed features attribute
        """

        self.fit(input_data)
        output_data = self.transform_for_fit(input_data)
        return output_data

    def _categorical_numerical_union(self, categorical_features: np.array, numerical_features: np.array) -> np.array:
        """Merge numerical and categorical features in right order (as it was in source table)
        """

        categorical_df = pd.DataFrame(categorical_features, columns=self.categorical_ids)
        numerical_df = pd.DataFrame(numerical_features, columns=self.non_categorical_ids)
        all_features_df = pd.concat([numerical_df, categorical_df], axis=1)

        # Sort column names
        all_features_df = all_features_df.sort_index(axis=1)
        return np.array(all_features_df)

    def _find_binary_features(self, numerical_features: np.array):
        """Find indices of features with only two unique values in column

        Notes:
            All features in table are numerical
        """

        df = pd.DataFrame(numerical_features)

        # Calculate unique values per column (excluding nans)
        for column_id, col in enumerate(df):
            unique_values = df[col].dropna().unique()
            if len(unique_values) <= 2:
                # Current numerical column has only two values
                column_info = {column_id: {'min': min(unique_values),
                                           'max': max(unique_values)}}
                self.ids_binary_integer_features.update(column_info)

    def _correct_binary_ids_features(self, filled_numerical_features: np.array) -> np.array:
        """ Correct filled features if previously it was binary. Discretization is performed
        for the reconstructed values

        Tip:
            [1, 1, 0.75, 0] will be transformed to [1, 1, 1, 0]
        """

        list_binary_ids = list(self.ids_binary_integer_features.keys())
        if len(list_binary_ids) == 0:
            # Return source array
            return filled_numerical_features

        for bin_id in list_binary_ids:
            # Correct values inplace
            filled_column = filled_numerical_features[:, bin_id]
            min_value = self.ids_binary_integer_features[bin_id]['min']
            max_value = self.ids_binary_integer_features[bin_id]['max']
            mean_value = (max_value - min_value) / 2

            filled_column[filled_column > mean_value] = max_value
            filled_column[filled_column < mean_value] = min_value

        return filled_numerical_features

    def get_params(self) -> OperationParameters:
        features_imputers = {'imputer_categorical': self.params_cat,
                             'imputer_numerical': self.params_num}
        return OperationParameters(**features_imputers)



from typing import Optional

import numpy as np
from sklearn.preprocessing import OneHotEncoder

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations. \
    implementation_interfaces import DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.tasks import Task, TaskTypesEnum


class DecomposerImplementation(DataOperationImplementation):
    """ Base class for decomposing target. The idea is to find the difference
    between the actual and predicted values - the residuals. Then the residuals
    replace the original target.
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    def fit(self, input_data: InputData):
        """
        The decompose operation doesn't support fit method
        """
        pass

    def transform(self, input_data: InputData) -> OutputData:
        """
        Method for modifying input_data for predict stage
        :param input_data: data with features, target and ids

        :return input_data: data with transformed features attribute
        """
        raise NotImplementedError()

    @staticmethod
    def divide_inputs(input_data: InputData) -> (np.array, np.array):
        """ Method for dividing InputData into parts:
        first came from Model parent and second came from Data parent

        :param input_data: InputData object
        :return prev_prediction: data obtained from "Model parent" at the previous node
        :return prev_features: data obtained from "Data parent" at the previous node
        """

        features = np.array(input_data.features)
        # Array with masks
        features_mask = np.array(input_data.supplementary_data.compound_mask)
        unique_features_masks = np.unique(features_mask)

        if len(unique_features_masks) < 2:
            prefix = 'Decompose operation must have at least two parents nodes'
            raise ValueError(f'{prefix}, but got {len(unique_features_masks)}')

        # Get amount of nodes data already visited
        flow_lengths = input_data.supplementary_data.flow_mask

        # Find minimum and maximum of visited nodes and first indices of them
        min_flow_length_i = np.argmin(flow_lengths)
        max_flow_length_i = np.argmax(flow_lengths)

        # For case when data from "Model parent" and "Data parent" go through equal number of nodes
        if min_flow_length_i == max_flow_length_i:
            # Find data models
            model_parent, data_parent = input_data.supplementary_data.define_parents(unique_features_masks,
                                                                                     task=input_data.task.task_type)
        else:
            model_parent = features_mask[max_flow_length_i]
            data_parent = features_mask[min_flow_length_i]

        # Get prediction from "Model parent"
        prev_prediction_id = np.ravel(np.argwhere(features_mask == model_parent))
        prev_prediction = features[:, prev_prediction_id]

        # Get prediction from "Data parent" - it must be the last parent in parent list
        prev_features_id = np.ravel(np.argwhere(features_mask == data_parent))
        prev_features = features[:, prev_features_id]

        return prev_prediction, prev_features

    def get_params(self) -> OperationParameters:
        return OperationParameters()


class DecomposerRegImplementation(DecomposerImplementation):
    """ Class for decomposing target for regression task """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    def transform(self, input_data: InputData) -> OutputData:
        """
        Method for modifying input_data for predict stage
        :param input_data: data with features, target and ids

        :return input_data: data with transformed features attribute
        """
        prev_prediction, prev_features = self.divide_inputs(input_data)
        output_data = self._get_output_data(input_data, prev_features)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """
        Method for modifying input_data for fit stage
        :param input_data: data with features, target and ids

        :return input_data: data with transformed features attribute
        """
        prev_prediction, prev_features = self.divide_inputs(input_data)

        # Target must be a column or table, not one-dimensional array
        target = np.array(input_data.target)
        if len(target.shape) < 2:
            target = target.reshape((-1, 1))

        diff = target - prev_prediction
        input_data.target = diff

        output_data = self._get_output_data(input_data, prev_features)
        return output_data

    def _get_output_data(self, input_data: InputData, prev_features: np.array) -> OutputData:
        output_data = self._convert_to_output(input_data, prev_features)
        # We decompose the target, so in the future we need to ignore
        output_data.supplementary_data.is_main_target = False
        return output_data


class DecomposerClassImplementation(DecomposerImplementation):
    """ Class for decomposing target for both binary and multiclass
    classification task
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    def transform(self, input_data: InputData) -> OutputData:
        """
        Method for modifying input_data for predict stage
        :param input_data: data with features, target and ids
        :return input_data: data with transformed features attribute
        """
        prev_prediction, prev_features = self.divide_inputs(input_data)
        output_data = self._get_output_data(input_data, prev_features)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """
        Method for modifying input_data for fit stage
        :param input_data: data with features, target and ids
        :return input_data: data with transformed features attribute
        """
        prev_prediction, prev_features = self.divide_inputs(input_data)

        target = np.array(input_data.target)
        if len(target.shape) < 2:
            target = target.reshape((-1, 1))

        input_data.target = self._get_difference(target, prev_prediction)
        output_data = self._get_output_data(input_data, prev_features)
        return output_data

    def _get_output_data(self, input_data: InputData, prev_features: np.array) -> OutputData:
        regression_task = Task(TaskTypesEnum.regression)
        output_data = self._convert_to_output(input_data, prev_features)
        output_data.supplementary_data.is_main_target = False
        output_data.task = regression_task
        return output_data

    def _get_difference(self, target: np.array, prev_prediction: np.array) -> np.array:
        """ Calculates difference between predictions (probabilities) and target
        :param target: class labels
        :param prev_prediction: predictions from previous classification model
        :return diff: difference between probabilities of classes
        """
        classes = np.unique(target)
        if len(classes) > 2:
            diff = self._multi_difference(target, prev_prediction)
        else:
            diff = self._binary_difference(classes, target, prev_prediction)
        return diff

    @staticmethod
    def _binary_difference(classes: np.array, target: np.array, prev_prediction: np.array) -> np.array:
        """ Calculates difference between predictions (probabilities) and target
        for binary classification task
        :param classes: which classes are in the target
        :param target: class labels
        :param prev_prediction: predictions from previous classification model
        :return diff: difference between probabilities of classes
        """
        minus_class = np.min(classes)
        plus_class = np.max(classes)

        minus_ids = np.argwhere(target == minus_class)
        plus_ids = np.argwhere(target == plus_class)

        # Replace class labels with probabilities (0.0 or 1.0)
        bin_target = np.copy(target)
        bin_target[minus_ids] = 0.0
        bin_target[plus_ids] = 1.0

        diff = bin_target - prev_prediction

        return diff

    @staticmethod
    def _multi_difference(target: np.array, prev_prediction: np.array) -> np.array:
        """ Calculates difference between predictions (probabilities) and target
        for multiclass classification task

        :param target: class labels
        :param prev_prediction: predictions from previous classification model
        :return diff: difference between probabilities of classes
        """

        # Make one-hot encoding for target
        binary_enc = OneHotEncoder().fit_transform(target)
        probabilities_target = binary_enc.toarray()
        diff = probabilities_target - prev_prediction

        return diff



import warnings

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_absolute_percentage_error

from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData

from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from copy import deepcopy
warnings.filterwarnings('ignore')


def prepare_input_data(len_forecast, train_data_features, train_data_target,
                       test_data_features):
    """ Function return prepared data for fit and predict

    :param len_forecast: forecast length
    :param train_data_features: time series which can be used as predictors for train
    :param train_data_target: time series which can be used as target for train
    :param test_data_features: time series which can be used as predictors for prediction

    :return train_input: Input Data for fit
    :return predict_input: Input Data for predict
    :return task: Time series forecasting task with parameters
    """

    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=len_forecast))

    train_input = InputData(idx=np.arange(0, len(train_data_features)),
                            features=train_data_features,
                            target=train_data_target,
                            task=task,
                            data_type=DataTypesEnum.ts)

    start_forecast = len(train_data_features)
    end_forecast = start_forecast + len_forecast
    predict_input = InputData(idx=np.arange(start_forecast, end_forecast),
                              features=test_data_features,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.ts)

    return train_input, predict_input, task


def get_arima_pipeline():
    """ Function return complex pipeline with the following structure
        arima
    """

    node_1 = PrimaryNode('arima')
    node_final = SecondaryNode('linear', nodes_from=[node_1])

    pipeline = Pipeline(node_final)
    return pipeline


def get_arima_nemo_pipeline():
    """ Function return complex pipeline with the following structure
        arima \
               linear
        nemo  |
    """

    node_arima = PrimaryNode('arima')
    node_nemo = PrimaryNode('exog_ts')
    node_final = SecondaryNode('ridge', nodes_from=[node_arima, node_nemo])
    pipeline = Pipeline(node_final)
    return pipeline


def return_working_pipeline():
    node_lagged_1 = PrimaryNode('lagged/1')
    node_exog = PrimaryNode('exog_ts')

    node_final = SecondaryNode('ridge', nodes_from=[node_lagged_1, node_exog])
    pipeline = Pipeline(node_final)
    return pipeline


len_forecast = 40
ts_name = 'sea_level'
path_to_file = '../../cases/data/nemo/sea_surface_height.csv'
path_to_exog_file = '../../cases/data/nemo/sea_surface_height_nemo.csv'

df = pd.read_csv(path_to_file)
time_series = np.array(df[ts_name])
df = pd.read_csv(path_to_exog_file)
exog_variable = np.array(df[ts_name])


# Let's divide our data on train and test samples
train_data = time_series[:-len_forecast]
test_data = time_series[-len_forecast:]

# Nemo features
train_data_exog = exog_variable[:-len_forecast]
test_data_exog = exog_variable[-len_forecast:]

# Source time series
train_input, predict_input, task = prepare_input_data(len_forecast=len_forecast,
                                                      train_data_features=train_data,
                                                      train_data_target=train_data,
                                                      test_data_features=train_data)

# Exogenous time series
train_input_exog, predict_input_exog, _ = prepare_input_data(len_forecast=len_forecast,
                                                             train_data_features=train_data_exog,
                                                             train_data_target=train_data,
                                                             test_data_features=test_data_exog)

pipeline = get_arima_pipeline()
train_dataset = MultiModalData({
        'arima': deepcopy(train_input),
    })
predict_dataset = MultiModalData({
        'arima': deepcopy(predict_input),
    })
pipeline.fit_from_scratch(train_dataset)
predicted_values = pipeline.predict(predict_dataset)
predicted_values = predicted_values.predict

predicted = np.ravel(np.array(predicted_values))
test_data = np.ravel(np.array(test_data))

mse_before = mean_squared_error(test_data, predicted, squared=False)
mae_before = mean_absolute_error(test_data, predicted)
mape_before = mean_absolute_percentage_error(test_data, predicted)
print(f'ARIMA MSE - {mse_before:.4f}')
print(f'ARIMA MAE - {mae_before:.4f}')
print(f'ARIMA MAPE - {mape_before:.4f}\n')


# arima with nemo ensemble
pipeline = return_working_pipeline()
train_dataset = MultiModalData({
        'lagged/1': deepcopy(train_input),
        'exog_ts': deepcopy(train_input_exog)
    })
predict_dataset = MultiModalData({
        'lagged/1': deepcopy(predict_input),
        'exog_ts': deepcopy(predict_input_exog)
    })
pipeline.fit_from_scratch(train_dataset)
predicted_values = pipeline.predict(predict_dataset).predict

predicted = np.ravel(np.array(predicted_values))
test_data = np.ravel(np.array(test_data))


mse_before = mean_squared_error(test_data, predicted, squared=False)
mae_before = mean_absolute_error(test_data, predicted)
mape_before = mean_absolute_percentage_error(test_data, predicted)
print(f'Lagged with nemo MSE - {mse_before:.4f}')
print(f'Lagged with nemo MAE - {mae_before:.4f}')
print(f'Lagged with nemo MAPE - {mape_before:.4f}\n')


# arima with nemo ensemble
pipeline = get_arima_nemo_pipeline()
train_dataset = MultiModalData({
        'arima': deepcopy(train_input),
        'exog_ts': deepcopy(train_input_exog)
    })
predict_dataset = MultiModalData({
        'arima': deepcopy(predict_input),
        'exog_ts': deepcopy(predict_input_exog)
    })
pipeline.fit_from_scratch(train_dataset)
predicted_values = pipeline.predict(predict_dataset).predict

predicted = np.ravel(np.array(predicted_values))
test_data = np.ravel(np.array(test_data))


mse_before = mean_squared_error(test_data, predicted, squared=False)
mae_before = mean_absolute_error(test_data, predicted)
mape_before = mean_absolute_percentage_error(test_data, predicted)
print(f'ARIMA with nemo MSE - {mse_before:.4f}')
print(f'ARIMA with nemo MAE - {mae_before:.4f}')
print(f'ARIMA with nemo MAPE - {mape_before:.4f}')



import numpy as np
from typing import Optional


class HyperparametersPreprocessor:
    """
    Class for hyperparameters preprocessing before operation fitting
    :param operation_type: name of the operation
    :param n_samples_data: number of rows in data
    """

    all_preprocessing_rules = {
        'knnreg': {
            'n_neighbors': ['integer']
        },
        'dtreg': {
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'treg': {
            'n_estimators': ['integer'],
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'rfr': {
            'n_estimators': ['integer'],
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'adareg': {
            'n_estimators': ['integer'],
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'gbr': {
            'n_estimators': ['integer'],
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'xgbreg': {
            'nthread': ['integer'],
            'n_estimators': ['integer'],
            'max_depth': ['integer'],
            'max_leaves': ['integer'],
            'max_bin': ['integer']
        },
        'lgbmreg': {
            'num_iterations': ['integer'], 'num_iteration': ['integer'], 'n_iter': ['integer'],
            'num_tree': ['integer'], 'num_trees': ['integer'],
            'num_round': ['integer'], 'num_rounds': ['integer'], 'nrounds': ['integer'],
            'num_boost_round': ['integer'], 'n_estimators': ['integer'], 'max_iter': ['integer'],
            'num_leaves': ['integer'], 'num_leaf': ['integer'],
            'max_leaves': ['integer'], 'max_leaf': ['integer'], 'max_leaf_nodes': ['integer'],
            'max_depth': ['integer'],
            'min_data_in_leaf': ['absolute', 'integer'], 'min_data_per_leaf': ['absolute', 'integer'],
            'min_data': ['absolute', 'integer'], 'min_child_samples': ['absolute', 'integer'],
            'min_samples_leaf': ['absolute', 'integer'],
            'bagging_freq': ['integer'], 'subsample_freq': ['integer'],
            'max_bin': ['integer'], 'max_bins': ['integer'],
        },
        'catboostreg': {
            'iterations': ['integer'], 'num_boost_round': ['integer'],
            'n_estimators': ['integer'], 'num_trees': ['integer'],
            'depth': ['integer'], 'max_depth': ['integer'],
            'min_data_in_leaf': ['absolute', 'integer'], 'min_child_samples': ['absolute', 'integer'],
            'max_leaves': ['integer'], 'num_leaves': ['integer'],
        },

        'knn': {
            'n_neighbors': ['integer']
        },
        'dt': {
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'rf': {
            'n_estimators': ['integer'],
            'max_leaf_nodes': ['le0_to_none', 'integer'],
            'max_depth': ['le0_to_none', 'integer']
        },
        'xgboost': {
            'nthread': ['integer'],
            'n_estimators': ['integer'],
            'max_depth': ['integer'],
            'max_leaves': ['integer'],
            'max_bin': ['integer']
        },
        'lgbm': {
            'num_iterations': ['integer'], 'num_iteration': ['integer'], 'n_iter': ['integer'],
            'num_tree': ['integer'], 'num_trees': ['integer'],
            'num_round': ['integer'], 'num_rounds': ['integer'], 'nrounds': ['integer'],
            'num_boost_round': ['integer'], 'n_estimators': ['integer'], 'max_iter': ['integer'],
            'num_leaves': ['integer'], 'num_leaf': ['integer'],
            'max_leaves': ['integer'], 'max_leaf': ['integer'], 'max_leaf_nodes': ['integer'],
            'max_depth': ['integer'],
            'min_data_in_leaf': ['absolute', 'integer'], 'min_data_per_leaf': ['absolute', 'integer'],
            'min_data': ['absolute', 'integer'], 'min_child_samples': ['absolute', 'integer'],
            'min_samples_leaf': ['absolute', 'integer'],
            'bagging_freq': ['integer'], 'subsample_freq': ['integer'],
            'max_bin': ['integer'], 'max_bins': ['integer'],
        },
        'catboost': {
            'iterations': ['integer'], 'num_boost_round': ['integer'],
            'n_estimators': ['integer'], 'num_trees': ['integer'],
            'depth': ['integer'], 'max_depth': ['integer'],
            'min_data_in_leaf': ['absolute', 'integer'], 'min_child_samples': ['absolute', 'integer'],
            'max_leaves': ['integer'], 'num_leaves': ['integer'],
        },

        'kmeans': {
            'n_clusters': ['integer']
        },

        'kernel_pca': {
            'n_components': ['le0_to_none', 'integer']
        },
        'fast_ica': {
            'n_components': ['le0_to_none', 'integer']
        },
    }

    def __init__(self,
                 operation_type: Optional[str],
                 n_samples_data: int):
        self.preprocessing_rules = self._get_preprocessing_rules(operation_type)
        self.n_samples_data = n_samples_data

        self.preprocessing_types_dict = {
            'integer': self._correct_integer,
            'absolute': self._correct_absolute,
            'le0_to_none': self._correct_le0_to_none
        }

    @staticmethod
    def _get_preprocessing_rules(operation_type):
        return HyperparametersPreprocessor.all_preprocessing_rules.get(operation_type, {})

    def correct(self,
                params: Optional[dict]) -> Optional[dict]:
        if params is None:
            return params

        for param in params:
            if param in self.preprocessing_rules:
                for preprocess in self.preprocessing_rules[param]:
                    params[param], is_final_transformation = self._correct(param_value=params[param],
                                                                           preprocess_type=preprocess)

                    if is_final_transformation:
                        break

        return params

    def _correct(self, param_value, preprocess_type):
        """
        Method calls preprocessing methods basing on preprocess type
        :param param_value : initial value of the parameter
        :param preprocess_type : type of the preprocessing transformation
        :return : param_value after preprocessing
        """

        if param_value is None:
            return None, True

        return self.preprocessing_types_dict[preprocess_type](param_value=param_value)

    def _correct_le0_to_none(self,
                             param_value):
        """
        Method converts all parameter values less or equal than 0 to None
        :param param_value: initial value of the parameter
        :return: (param_value after transformation, True if no more transformations needed)
        """

        if param_value <= 0:
            # None value do need to be transformed anymore
            return None, True
        return param_value, False

    def _correct_absolute(self,
                          param_value):
        """
        Method adds option of using share of total samples in data besides an absolute number of samples
        :param param_value: initial value of the parameter
        :return: param_value after transformation
        """

        if 0 <= param_value < 1:
            # Transformed to absolute value share do not need to be transformed anymore
            return int(np.ceil(param_value * self.n_samples_data)), True
        return param_value, False

    def _correct_integer(self,
                         param_value):
        """
        Method rounds parameter value to avoid errors
        :param param_value: initial value of the parameter
        :return: param_value after rounding
        """

        return round(param_value), False



from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline


def cnn_composite_pipeline(composite_flag: bool = True) -> Pipeline:
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/cnn_composite_pipeline.png
      :width: 55%

    Where cnn - convolutional neural network, rf - random forest

    :param composite_flag:  add additional random forest estimator
    """
    node_first = PrimaryNode('cnn')
    node_first.parameters = {'architecture': 'deep',
                             'epochs': 15,
                             'batch_size': 128}
    node_second = PrimaryNode('cnn')
    node_second.parameters = {'architecture_type': 'simplified',
                              'epochs': 10,
                              'batch_size': 128}
    node_final = SecondaryNode('rf', nodes_from=[node_first, node_second])

    if not composite_flag:
        node_final = SecondaryNode('rf', nodes_from=[node_first])

    pipeline = Pipeline(node_final)
    return pipeline


def classification_pipeline_with_balancing(custom_params=None):
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/class_with_balancing.png
      :width: 55%

    Where resample - algorithm for balancing dataset, logit - logistic_regression

    :param custom_params:  custom parameters for resample node
    """
    node_resample = PrimaryNode(operation_type='resample')

    if custom_params is not None:
        node_resample.parameters = custom_params

    graph = SecondaryNode(operation_type='logit', nodes_from=[node_resample])

    return Pipeline(graph)


def classification_pipeline_without_balancing():
    """
    Returns: pipeline with the following structure:

    .. image:: img_classification_pipelines/class_without_balancing.png
      :width: 55%

    Where logit - logistic_regression
    """
    node = PrimaryNode(operation_type='logit')

    return Pipeline(node)


def classification_complex_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/complex_pipeline.png
      :width: 55%

    """
    first = PrimaryNode(operation_type='rf')
    second = PrimaryNode(operation_type='knn')
    final = SecondaryNode(operation_type='logit',
                          nodes_from=[first, second])

    pipeline = Pipeline(final)

    return pipeline


def classification_random_forest_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/random_forest.png
      :width: 55%

    """
    node_scaling = PrimaryNode('scaling')
    node_final = SecondaryNode('rf', nodes_from=[node_scaling])
    return Pipeline(node_final)


def classification_isolation_forest_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/isolation_forest.png
      :width: 55%

    """
    node_first = PrimaryNode('scaling')
    node_second = SecondaryNode('isolation_forest_class', nodes_from=[node_first])
    node_final = SecondaryNode('rf', nodes_from=[node_second])
    return Pipeline(node_final)


def classification_svc_complex_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/svc_complex_pipeline.png
      :width: 55%

    Where svc - support vector classifier, logit - logistic regression, knn - K nearest neighbors classifier,
    rf - random forest classifier

    """

    svc_primary_node = PrimaryNode('svc')
    svc_primary_node.parameters = dict(probability=True)
    logit_secondary_node = SecondaryNode('logit', nodes_from=[svc_primary_node])

    svc_node_with_custom_params = PrimaryNode('svc')
    svc_node_with_custom_params.parameters = dict(kernel='rbf', C=10,
                                                     gamma=1, cache_size=2000,
                                                     probability=True)
    logit_secondary_node_2 = SecondaryNode('logit', nodes_from=[svc_node_with_custom_params])

    knn_primary_node = PrimaryNode('knn')
    knn_secondary_node = SecondaryNode('knn', nodes_from=[knn_primary_node, logit_secondary_node])

    rf_node = SecondaryNode('rf', nodes_from=[logit_secondary_node_2, knn_secondary_node])

    preset_pipeline = Pipeline(rf_node)

    return preset_pipeline


def classification_three_depth_manual_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/manual_three_depth_pipeline.png
      :width: 55%

    Where rf - xg boost classifier, logit - logistic regression, knn - K nearest neighbors classifier,
    qda - discriminant analysis
   """
    logit_node_primary = PrimaryNode('logit')
    xgb_node_primary = PrimaryNode('rf')
    xgb_node_primary_second = PrimaryNode('rf')

    qda_node_third = SecondaryNode('qda', nodes_from=[xgb_node_primary_second])
    knn_node_third = SecondaryNode('knn', nodes_from=[logit_node_primary, xgb_node_primary])

    knn_root = SecondaryNode('knn', nodes_from=[qda_node_third, knn_node_third])

    pipeline = Pipeline(knn_root)

    return pipeline


def classification_rf_complex_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_classification_pipelines/complex_rf_pipeline.png
      :width: 55%

    Where lda - discriminant analysis, logit - logistic regression, rf - random forest classifier,
    knn - K nearest neighbors classifier
    """
    pipeline = Pipeline()

    root_of_tree, root_child_first, root_child_second = \
        [SecondaryNode(model) for model in ('rf', 'rf', 'knn')]

    for root_node_child in (root_child_first, root_child_second):
        for requirement_model in ('logit', 'lda'):
            new_node = PrimaryNode(requirement_model)
            root_node_child.nodes_from.append(new_node)
            pipeline.add_node(new_node)
        pipeline.add_node(root_node_child)
        root_of_tree.nodes_from.append(root_node_child)

    pipeline.add_node(root_of_tree)
    return pipeline



from typing import Optional

import numpy as np
from sklearn.multiclass import OneVsRestClassifier
from sklearn.svm import SVC

from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class FedotSVCImplementation(ModelImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        if not self.params:
            default_params = {'kernel': 'linear', 'probability': True, 'class_weight': 'balanced'}
            self.params.update(**default_params)
        self.inner_model = SVC(**self.params.to_dict())
        self.model = OneVsRestClassifier(self.inner_model)
        self.classes = None

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """
        self.classes = np.unique(train_data.target)
        self.model.fit(train_data.features, train_data.target)
        return self.model

    def predict(self, input_data):
        """ Method make prediction with labels of classes

        :param input_data: data with features to process
        """
        prediction = self.model.predict(input_data.features)

        return prediction

    def predict_proba(self, input_data):
        """ Method make prediction with probabilities of classes

        :param input_data: data with features to process
        """
        prediction = self.model.predict_proba(input_data.features)

        return prediction

    @property
    def classes_(self):
        return self.classes



import json
import os
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import numpy as np

from fedot.core.constants import BEST_QUALITY_PRESET_NAME, AUTO_PRESET_NAME
from fedot.core.log import default_log
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.json_evaluation import eval_field_str, eval_strategy_str, read_field
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence

AVAILABLE_REPO_NAMES = ['all', 'model', 'data_operation', 'automl']


@dataclass
class OperationMetaInfo:
    id: str
    input_types: List[DataTypesEnum]
    output_types: List[DataTypesEnum]
    task_type: List[TaskTypesEnum]
    supported_strategies: Any
    allowed_positions: List[str]
    tags: Optional[List[str]] = None
    presets: Optional[List[str]] = None

    def current_strategy(self, task: TaskTypesEnum):
        """Method allows getting available processing strategies depending on the
        selected task

        Args:
            task: machine learning task (e.g. regression and classification)

        Returns:
            supported strategies for task
        """

        if isinstance(self.supported_strategies, dict):
            return self.supported_strategies.get(task, None)
        return self.supported_strategies


def run_once(function):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return function(*args, **kwargs)

    wrapper.has_run = False
    return wrapper


class OperationTypesRepository:
    """Class for connecting models and data operations with json files with
    its descriptions and metadata
    """

    __initialized_repositories__ = {}
    # The later the tag, the higher its priority in case of intersection
    DEFAULT_MODEL_TAGS = ['linear', 'non_linear', 'custom_model', 'tree', 'boosting', 'ts_model', 'deep']
    DEFAULT_DATA_OPERATION_TAGS = [
        'data_source', 'feature_scaling', 'imputation', 'feature_reduction', 'feature_engineering', 'encoding',
        'filtering', 'feature_selection', 'ts_to_table', 'smoothing', 'ts_to_ts', 'text', 'decompose', 'imbalanced',
        'data_source_img', 'data_source_text', 'data_source_table', 'data_source_ts'
    ]

    __repository_dict__ = {
        'model': {'file': 'model_repository.json', 'initialized_repo': None, 'default_tags': DEFAULT_MODEL_TAGS},
        'data_operation': {'file': 'data_operation_repository.json', 'initialized_repo': None,
                           'default_tags': DEFAULT_DATA_OPERATION_TAGS},
        'automl': {'file': 'automl_repository.json', 'initialized_repo': None, 'default_tags': []}

    }

    def __init__(self, operation_type: str = 'model'):
        self.log = default_log(self)

        self._tags_excluded_by_default = ['non-default', 'expensive']
        OperationTypesRepository.init_default_repositories()

        self.operation_type = operation_type
        self.repository_name = []
        self._repo = []
        self.default_tags = []
        if operation_type == 'all':
            for op_type in OperationTypesRepository.__repository_dict__.keys():
                self.repository_name.append(OperationTypesRepository.__repository_dict__[op_type]['file'])
                operations = OperationTypesRepository.__repository_dict__[op_type]['initialized_repo']
                self.default_tags += OperationTypesRepository.__repository_dict__[op_type]['default_tags']

                if operations is not None:
                    for operation in operations:
                        if operation not in self._repo:
                            self._repo.append(operation)

        else:
            self.repository_name = OperationTypesRepository.__repository_dict__[operation_type]['file']
            self._repo = OperationTypesRepository.__repository_dict__[operation_type]['initialized_repo']
            self.default_tags = OperationTypesRepository.__repository_dict__[operation_type]['default_tags']

    @classmethod
    def get_available_repositories(cls):
        operation_types = []
        for t in cls.__repository_dict__:
            if cls.__repository_dict__[t]['initialized_repo'] is not None:
                operation_types.append(t)
        return operation_types

    @classmethod
    def init_automl_repository(cls):
        default_automl_repo_file = cls.__repository_dict__['automl']['file']
        return cls.assign_repo('automl', default_automl_repo_file)

    @classmethod
    @run_once
    def init_default_repositories(cls):
        # default model repo
        default_model_repo_file = cls.__repository_dict__['model']['file']
        cls.assign_repo('model', default_model_repo_file)

        # default data_operation repo
        default_data_operation_repo_file = cls.__repository_dict__['data_operation']['file']
        cls.assign_repo('data_operation', default_data_operation_repo_file)

    @classmethod
    def assign_repo(cls, operation_type: str, repo_file: str):
        if operation_type not in cls.__repository_dict__:
            raise Warning(f'The {operation_type} is not supported. The model type will be set')

        repo_path = create_repository_path(repo_file)
        if repo_file not in cls.__initialized_repositories__.keys():
            cls.__initialized_repositories__[repo_file] = cls._initialise_repo(repo_path)
        cls.__repository_dict__[operation_type]['file'] = repo_file
        cls.__repository_dict__[operation_type]['initialized_repo'] = cls.__initialized_repositories__[repo_file]

        return OperationTypesRepository(operation_type)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.repo_path = None
        OperationTypesRepository.__repository_dict__[self.operation_type]['initialized_repo'] = None
        default_model_repo_file = OperationTypesRepository.__repository_dict__['model']['file']
        OperationTypesRepository.assign_repo('model', default_model_repo_file)

    def __repr__(self):
        return f"{self.__class__.__name__} for {self.repository_name}"

    @property
    def is_initialized(self):
        return OperationTypesRepository.__repository_dict__[self.operation_type]['initialized_repo'] is None

    @classmethod
    def _initialise_repo(cls, repo_path: str) -> List[OperationMetaInfo]:
        """Method parse ``JSON`` repository with operations descriptions and
        wrapped information into :obj:`OperationMetaInfo`, then put it into the list

        Returns:
            List[OperationMetaInfo]: list with :obj:`OperationMetaInfo` for every operation from ``json`` repository
        """

        repository_json = load_repository(repo_path)

        metadata_json = repository_json['metadata']
        operations_json = repository_json['operations']

        operations_list = []
        for current_operation_key in operations_json:
            # Get information about operation
            # properties - information about operation by key, for example tags
            # metadata - information about meta of the operation
            properties = operations_json.get(current_operation_key)
            metadata = metadata_json[properties['meta']]

            task_types = eval_field_str(metadata['tasks'])
            input_type = eval_field_str(properties['input_type']) \
                if ('input_type' in properties) \
                else eval_field_str(metadata['input_type'])
            output_type = eval_field_str(properties['output_type']) \
                if ('output_type' in properties) \
                else eval_field_str(metadata['output_type'])

            # Get available strategies for obtained metadata
            supported_strategies = OperationTypesRepository.get_strategies_by_metadata(metadata)

            accepted_node_types = read_field(metadata, 'accepted_node_types', ['any'])
            forbidden_node_types = read_field(metadata, 'forbidden_node_types', [])

            # Get tags for meta and for operation
            meta_tags = read_field(metadata, 'tags', [])
            operation_tags = read_field(properties, 'tags', [])
            presets = read_field(properties, 'presets', [])

            allowed_positions = ['primary', 'secondary', 'root']

            if accepted_node_types and accepted_node_types != 'all':
                allowed_positions = accepted_node_types
            if forbidden_node_types:
                allowed_positions = [pos for pos in allowed_positions if
                                     pos not in forbidden_node_types]

            # Unit tags
            tags = meta_tags + operation_tags

            operation = OperationMetaInfo(id=current_operation_key,
                                          input_types=input_type,
                                          output_types=output_type,
                                          task_type=task_types,
                                          supported_strategies=supported_strategies,
                                          allowed_positions=allowed_positions,
                                          tags=tags,
                                          presets=presets)
            operations_list.append(operation)

        return operations_list

    @staticmethod
    def get_strategies_by_metadata(metadata: dict):
        """Method allow obtain strategy instance by the metadata

        Args:
            metadata: information about meta of the operation
            supported_strategies: available strategies for current metadata
        """
        strategies_json = metadata['strategies']
        if isinstance(strategies_json, list):
            supported_strategies = eval_strategy_str(strategies_json)
        else:
            supported_strategies = {}
            for strategy_dict_key in strategies_json.keys():
                # Convert string into class path for import
                import_path = eval_field_str(strategy_dict_key)
                strategy_class = eval_strategy_str(strategies_json[strategy_dict_key])

                supported_strategies.update({import_path: strategy_class})
        return supported_strategies

    def operation_info_by_id(self, operation_id: str) -> Optional[OperationMetaInfo]:
        """ Get operation by it's name (id) """

        operation_id = get_operation_type_from_id(operation_id)

        operations_with_id = [m for m in self._repo if m.id == operation_id]
        if len(operations_with_id) > 1:
            raise ValueError('Several operations with same id in repository')
        if len(operations_with_id) == 0:
            self.log.warning(f'Operation {operation_id} not found in the repository')
            return None
        return operations_with_id[0]

    def operations_with_tag(self, tags: List[str], is_full_match: bool = False):
        """ Method returns operations from repository with specific tags

                :param tags: list of required tags
                :param is_full_match: are all tags are required

                :return list of suitable operations names
        """
        operations_info = [m for m in self._repo if
                           _is_operation_contains_tag(tags, m.tags, is_full_match)]
        return [m.id for m in operations_info]

    def suitable_operation(self, task_type: TaskTypesEnum = None,
                           data_type: DataTypesEnum = None,
                           tags: List[str] = None, is_full_match: bool = False,
                           forbidden_tags: List[str] = None,
                           preset: str = None) -> List[str]:
        """Method returns operations from repository for desired task and / or
        tags. Filtering method.

        Args:
            task_type: task to filter
            data_type: data type to filter
            tags: operations with which tags are required
            is_full_match: requires all tags to match, or at least one
            forbidden_tags: operations with such tags shouldn't be returned
            preset: return operations from desired preset
        """

        if not forbidden_tags:
            forbidden_tags = []

        if not tags:
            for excluded_default_tag in self._tags_excluded_by_default:
                # Forbidden tags by default
                forbidden_tags.append(excluded_default_tag)

        no_task = task_type is None
        operations_info = []
        for o in self._repo:
            is_desired_task = task_type in o.task_type or no_task
            tags_good = not tags or _is_operation_contains_tag(tags, o.tags, is_full_match)
            tags_bad = not forbidden_tags or not _is_operation_contains_tag(forbidden_tags, o.tags, False)
            is_desired_preset = _is_operation_contains_preset(o.presets, preset)
            if is_desired_task and tags_good and tags_bad and is_desired_preset:
                operations_info.append(o)

        if data_type:
            # ignore text and image data types: there are no operations with these `input_type`
            ignore_data_type = data_type in [DataTypesEnum.text, DataTypesEnum.image]
            if data_type == DataTypesEnum.ts:
                valid_data_types = [DataTypesEnum.ts, DataTypesEnum.table]
            else:
                valid_data_types = ensure_wrapped_in_sequence(data_type)
            if not ignore_data_type:
                operations_info = [o for o in operations_info if
                                   np.any([data_type in o.input_types for data_type in valid_data_types])]

        return [m.id for m in operations_info]

    @property
    def operations(self):
        return self._repo

    def get_first_suitable_operation_tag(self, operation: str, tags_to_find: Optional[List[str]] = None) \
            -> Optional[str]:
        """Finds the first suitable tag for the operation in the repository.

        Args:
            operation: name of the operation
            tags_to_find: list of suitable tags. The later the tag, the higher its priority in case of intersection

        Returns:
            Optional[List[str]]: first suitable tag or ``None``
        """

        tags_to_find = tags_to_find or self.default_tags

        info = self.operation_info_by_id(operation)
        if info is None:
            return None
        for tag in reversed(tags_to_find):
            if tag in info.tags:
                return tag
        return None


def get_visualization_tags_map() -> Dict[str, List[str]]:
    """
    Returns map between repository tags and list of corresponding models for visualizations.
    """
    # Search for tags.
    operations_map = {}
    for repo_name in ('model', 'data_operation'):
        repo = OperationTypesRepository(repo_name)
        for operation in repo.operations:
            tag = repo.get_first_suitable_operation_tag(operation.id, repo.default_tags)
            operations_map[tag] = (operations_map.get(tag) or []) + [operation.id]
    # Sort tags.
    tags_model = OperationTypesRepository.DEFAULT_MODEL_TAGS
    tags_data = OperationTypesRepository.DEFAULT_DATA_OPERATION_TAGS
    operations_map = {tag: operations_map[tag] for tag in tags_model + tags_data if tag in operations_map}
    return operations_map


def _is_operation_contains_tag(candidate_tags: List[str],
                               operation_tags: List[str],
                               is_full_match: bool) -> bool:
    """The function checks which operations are suitable for the selected tags

    Args:
        candidate_tags: ``list`` with tags that the operation must have in order
            to fit the selected task
        operation_tags: ``list`` with tags with names as in repository json file
            which correspond to the considering operation
        is_full_match: requires all tags to match, or at least one

    Returns:
        bool: is there a match on the tags
    """

    matches = (tag in operation_tags for tag in candidate_tags)
    if is_full_match:
        return all(matches)
    else:
        return any(matches)


def _is_operation_contains_preset(operation_presets: List[str], preset: str) -> bool:
    """Checking whether the operation is suitable for current preset
    """

    if preset is None:
        # None means that best_quality preset are using so return all operations
        return True

    return preset in operation_presets


def atomized_model_type():
    return 'atomized_operation'


def atomized_model_meta_tags():
    return ['random'], ['any'], ['atomized']


def get_operations_for_task(task: Optional[Task], data_type: Optional[DataTypesEnum] = None, mode='all', tags=None,
                            forbidden_tags=None, preset: str = None):
    """Function returns aliases of operations.

    Args:
        task: task to solve
        data_type: type of input data
        mode: mode to return operations

            .. details:: the possible parameters of ``mode``:

                - ``all`` -> return list with all operations
                - ``model`` -> return only list with models
                - ``data_operation`` -> return only list with data_operations

        tags: tags for grabbing when filtering
        forbidden_tags: tags for skipping when filtering
        preset: operations from this preset will be obtained

    Returns:
        list:  operation aliases
    """

    # Preset None means that all operations will be returned
    if preset is not None:
        if BEST_QUALITY_PRESET_NAME in preset or AUTO_PRESET_NAME in preset:
            preset = None

    task_type = task.task_type if task else None
    if mode in AVAILABLE_REPO_NAMES:
        repo = OperationTypesRepository(mode)
        model_types = repo.suitable_operation(task_type, data_type=data_type, tags=tags, forbidden_tags=forbidden_tags,
                                              preset=preset)
        return model_types
    else:
        raise ValueError(f'Such mode "{mode}" is not supported')


def get_operation_type_from_id(operation_id):
    operation_type = _operation_name_without_postfix(operation_id)
    return operation_type


def _operation_name_without_postfix(operation_id):
    """
    Args:
        operation_id: operation name with optional postfix - text after / sign
    Returns:
        operation type - all characters before postfix (all characters if no postfix found)
    """

    postfix_sign = '/'
    # if the operation id has custom postfix
    if postfix_sign in operation_id:
        if operation_id.count(postfix_sign) > 1:
            raise ValueError(f'Incorrect number of postfixes in {operation_id}')
        return operation_id.split('/')[0]
    else:
        return operation_id


def load_repository(repo_path: str) -> dict:
    # Loads the repository for various cases and loads the necessary additional data "base_repository.json".
    with open(repo_path) as repository_json_file:
        repository_json = json.load(repository_json_file)

    if 'base_repository' in repository_json:
        base_repository_json_file = create_repository_path(repository_json['base_repository'])

        with open(base_repository_json_file) as repository_json_file:
            base_repository_json = json.load(repository_json_file)

        merged_dict = defaultdict(dict)

        merged_dict.update(base_repository_json)
        for key, nested_dict in repository_json.items():
            if key not in merged_dict:
                merged_dict[key] = nested_dict
            else:
                merged_dict[key].update(nested_dict)

        repository_json = dict(merged_dict)

    return repository_json


def create_repository_path(repository_name: str) -> str:
    # Path till repository file
    file = os.path.join('data', repository_name)
    # Path till current file with script
    repo_folder_path = str(os.path.dirname(__file__))
    repo_path = os.path.join(repo_folder_path, file)

    return repo_path



from copy import copy
from typing import Optional

import numpy as np
from sklearn.utils import resample

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import default_log
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import (
    DataOperationImplementation
)
from fedot.core.operations.operation_parameters import OperationParameters

GLOBAL_PREFIX = 'sklearn_imbalanced_class:'


# TODO: ResampleImplementation to multi-class imbalanced data


class ResampleImplementation(DataOperationImplementation):
    """Implementation of imbalanced bin class transformation for
    classification task by using method from sklearn.utils.resample

    Args:
        params: OperationParameters with the hyperparameters:
            balance: Data transformation strategy. Balance strategy can be 'expand_minority' or 'reduce_majority'.
                In case of expand_minority elements of minor class are expanding to n_samples.
                In otherwise with reduce_majority elements of major class are reducing to n_samples.
            replace: Implements resampling with replacement. If False, this will implement (sliced) random permutations.
            balance_ratio: Transformation ratio can take values in the range [0, 1].
                With balance_ratio = 0 nothing happens and data will remain the same.
                In case of balance_ratio = 1 means that both classes will be balanced and the shape of both will become
                equal. If balance_ratio < 1.0 means that the data of one class is getting closer to the shape of opposite
                class. If None numbers of samples will be equal to the shape of opposite selected transformed class.
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.n_samples = None
        self.log = default_log(self)

    @property
    def balance(self) -> str:
        return self.params.setdefault('balance', 'expand_minority')

    @property
    def replace(self) -> bool:
        return self.params.setdefault('replace', False)

    @property
    def balance_ratio(self) -> float:
        return self.params.setdefault('balance_ratio', 1.0)

    def fit(self, input_data: Optional[InputData]):
        """Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """
        pass

    def transform(self, input_data: InputData) -> OutputData:
        """Transformed input data via selected balance strategy for predict stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output_data: transformed input_data via strategy
        """
        output_data = self._convert_to_output(input_data, input_data.features,
                                              data_type=input_data.data_type)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """Transformed input data via selected balance strategy for fit stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output_data: transformed input_data via strategy
        """
        copied_data = copy(input_data)

        if len(np.unique(copied_data.target)) != 2:
            # Imbalanced multi-class balancing is not supported.
            return self._convert_to_output(input_data, input_data.features)

        unique_class, number_of_elements = np.unique(copied_data.target, return_counts=True)

        if number_of_elements[0] == number_of_elements[1]:
            # If number of elements of each class are equal that transformation is not required
            return self._convert_to_output(input_data, input_data.features)

        min_data, maj_data = self._get_data_by_target(copied_data.features, copied_data.target,
                                                      unique_class, number_of_elements)

        self._check_and_correct_balance_ratio_param()
        self.n_samples = self._convert_to_absolute(min_data, maj_data)
        self._check_and_correct_replace_param(min_data, maj_data)

        if self.balance == 'expand_minority':
            prev_shape = min_data.shape
            min_data = self._resample_data(min_data)
            self.log.debug(
                f'{GLOBAL_PREFIX} According to {self.balance} data was changed from {prev_shape} to {min_data.shape}'
            )

        elif self.balance == 'reduce_majority':
            prev_shape = maj_data.shape
            maj_data = self._resample_data(maj_data)
            self.log.debug(
                f'{GLOBAL_PREFIX} According to {self.balance} data was changed from {prev_shape} to {maj_data.shape}'
            )

        transformed_data = np.concatenate((min_data, maj_data), axis=0).transpose()

        output_data = OutputData(
            idx=np.arange(transformed_data.shape[1]),
            features=input_data.features,
            predict=transformed_data[:-1].transpose(),
            task=input_data.task,
            target=transformed_data[-1],
            data_type=input_data.data_type,
            supplementary_data=input_data.supplementary_data)
        return output_data

    @staticmethod
    def _get_data_by_target(features: np.array, target: np.array, unique: np.array,
                            number_of_elements: np.array) -> np.array:
        """Unify features and target in one array and split into classes
        """
        if number_of_elements[0] < number_of_elements[1]:
            min_idx = np.where(target == unique[0])[0]
            maj_idx = np.where(target == unique[1])[0]
        else:
            min_idx = np.where(target == unique[1])[0]
            maj_idx = np.where(target == unique[0])[0]

        minority_data = np.hstack((features[min_idx], target[min_idx].reshape(-1, 1)))
        majority_data = np.hstack((features[maj_idx], target[maj_idx].reshape(-1, 1)))

        return minority_data, majority_data

    def _check_and_correct_replace_param(self, min_data: np.array, maj_data: np.array):
        """Method checks if selected replace parameter is correct

        Args:
            min_data: minority data from input data
            maj_data: majority data from input data
        """
        if self.replace is False:
            if self.balance == 'expand_minority' and self.n_samples >= min_data.shape[0]:
                self.params.update(replace=True)
            elif self.balance == 'reduce_majority' and self.n_samples >= maj_data.shape[0]:
                self.params.update(replace=True)
            self.log.debug(f'{GLOBAL_PREFIX} resample operation allow repeats in data')

    def _check_and_correct_balance_ratio_param(self):
        """Method checks if selected balance_ratio parameter is correct
        """
        if not self.balance_ratio:
            self.params.update(balance_ratio=1)
        if self.balance_ratio < 0 or self.balance_ratio > 1:
            self.params.update(balance_ratio=1)
            self.log.debug(f'{GLOBAL_PREFIX} balance ratio set to full balance')

    def _convert_to_absolute(self, min_data: np.array, maj_data: np.array) -> float:
        self.log.debug(f'{GLOBAL_PREFIX} set n_samples in absolute values due to balance_ratio')
        difference = maj_data.shape[0] - min_data.shape[0]

        if self.balance == 'expand_minority':
            return round(difference * self.balance_ratio + min_data.shape[0])

        elif self.balance == 'reduce_majority':
            return round(maj_data.shape[0] - difference * self.balance_ratio)

    def _resample_data(self, data: np.array):
        return resample(data, replace=self.replace, n_samples=self.n_samples)



from random import choice
from typing import Optional

from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptNode
from fedot.core.optimisers.opt_node_factory import OptNodeFactory
from fedot.core.pipelines.pipeline_advisor import PipelineChangeAdvisor, check_for_specific_operations
from fedot.core.repository.pipeline_operation_repository import PipelineOperationRepository


class PipelineOptNodeFactory(OptNodeFactory):
    def __init__(self, requirements: PipelineComposerRequirements,
                 advisor: Optional[PipelineChangeAdvisor] = None,
                 graph_model_repository: Optional[PipelineOperationRepository] = None):
        self.requirements = requirements
        self.advisor = advisor or PipelineChangeAdvisor()
        self.graph_model_repository = graph_model_repository or self._init_default_graph_model_repo()

    def _init_default_graph_model_repo(self):
        """ Initialize default graph model repository with operations from composer requirements """
        repo = PipelineOperationRepository(operations_by_keys={'primary': self.requirements.primary,
                                                               'secondary': self.requirements.secondary})
        return repo

    def exchange_node(self,
                      node: OptNode):
        candidates = self.graph_model_repository.get_operations(is_primary=False) \
            if node.nodes_from else self.graph_model_repository.get_operations(is_primary=True)
        candidates = self.advisor.propose_change(node=node,
                                                 possible_operations=candidates)
        candidates = self.filter_specific_candidates(candidates)
        return self._return_node(candidates)

    def get_parent_node(self, node: OptNode, is_primary: bool):
        possible_operations = self.graph_model_repository.get_operations(is_primary=is_primary)
        candidates = self.advisor.propose_parent(node=node,
                                                 possible_operations=possible_operations)
        candidates = self.filter_specific_candidates(candidates)
        return self._return_node(candidates)

    def get_node(self,
                 is_primary: bool):
        candidates = self.graph_model_repository.get_operations(is_primary=is_primary)
        candidates = self.filter_specific_candidates(candidates)
        return self._return_node(candidates)

    @staticmethod
    def _return_node(candidates) -> Optional[OptNode]:
        if not candidates:
            return None
        return OptNode(content={'name': choice(candidates)})

    def filter_specific_candidates(self, candidates: list):
        return list(filter(lambda x: not check_for_specific_operations(x), candidates))



from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.inheritance import GeneticSchemeTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.operation_types_repository import get_operations_for_task


def get_three_depth_manual_class_pipeline():
    logit_node_primary = PrimaryNode('logit')
    xgb_node_primary = PrimaryNode('xgboost')
    xgb_node_primary_second = PrimaryNode('xgboost')

    qda_node_third = SecondaryNode('qda', nodes_from=[xgb_node_primary_second])
    knn_node_third = SecondaryNode('knn', nodes_from=[logit_node_primary, xgb_node_primary])

    knn_root = SecondaryNode('knn', nodes_from=[qda_node_third, knn_node_third])

    pipeline = Pipeline(knn_root)

    return pipeline


def get_three_depth_manual_regr_pipeline():
    rfr_primary = PrimaryNode('rfr')
    knn_primary = PrimaryNode('knnreg')

    dtreg_secondary = SecondaryNode('dtreg', nodes_from=[rfr_primary])
    rfr_secondary = SecondaryNode('rfr', nodes_from=[knn_primary])

    knnreg_root = SecondaryNode('knnreg', nodes_from=[dtreg_secondary, rfr_secondary])

    pipeline = Pipeline(knnreg_root)

    return pipeline


def get_composed_pipeline(dataset_to_compose, task, metric_function):
    # the search of the models provided by the framework that can be used as nodes in a pipeline for the selected task
    available_model_types = get_operations_for_task(task=task, mode='model')

    # the choice and initialisation of the GP search
    composer_requirements = PipelineComposerRequirements(
        primary=available_model_types,
        secondary=available_model_types,
    )
    params = GPGraphOptimizerParameters(
        genetic_scheme_type=GeneticSchemeTypesEnum.steady_state
    )

    # Create composer and with required composer params
    composer = ComposerBuilder(task=task). \
        with_requirements(composer_requirements). \
        with_optimizer_params(params). \
        with_metrics(metric_function). \
        build()

    # the optimal pipeline generation by composition - the most time-consuming task
    pipeline_evo_composed = composer.compose_pipeline(data=dataset_to_compose)

    return pipeline_evo_composed


def pipeline_by_task(task, metric, data, is_composed):
    if is_composed:
        pipeline = get_composed_pipeline(data, task,
                                         metric_function=metric)
    else:
        if task.task_type.name == 'classification':
            pipeline = get_three_depth_manual_class_pipeline()
        else:
            pipeline = get_three_depth_manual_regr_pipeline()

    return pipeline
